from typing import TYPE_CHECKING

from fastapi.testclient import TestClient

from testing.mock import MockSettings
from yak_server.helpers.settings import get_settings

if TYPE_CHECKING:
    from fastapi import FastAPI


def test_no_schema_introspection_in_production(production_app_session: "FastAPI") -> None:
    client = TestClient(production_app_session)

    production_app_session.dependency_overrides[get_settings] = MockSettings()

    response = client.post(
        "/api/v2",
        json={
            "query": """{
                __schema {
                    queryType {
                        name
                    }
                }
            }
            """,
        },
    )

    assert response.json() == {
        "data": None,
        "errors": [
            {
                "message": (
                    "GraphQL introspection has been disabled, "
                    "but the requested query contained the field '__schema'."
                ),
                "locations": [
                    {
                        "line": 2,
                        "column": 17,
                    },
                ],
            },
            {
                "message": (
                    "GraphQL introspection has been disabled, "
                    "but the requested query contained the field 'queryType'."
                ),
                "locations": [
                    {
                        "line": 3,
                        "column": 21,
                    },
                ],
            },
        ],
    }



from typing import TYPE_CHECKING, Optional

from pydantic import UUID4, BaseModel, ConfigDict, PositiveInt

from yak_server.helpers.language import Lang, get_language_description

from .groups import GroupIn, GroupOut
from .phases import PhaseOut
from .teams import FlagOut, TeamIn, TeamModifyScoreBetIn, TeamWithScoreOut

if TYPE_CHECKING:
    from yak_server.database.models import ScoreBetModel


class ScoreBetIn(BaseModel):
    index: PositiveInt
    team1: TeamIn
    team2: TeamIn
    group: GroupIn


class ScoreBetOut(BaseModel):
    id: UUID4
    locked: bool
    team1: Optional[TeamWithScoreOut] = None
    team2: Optional[TeamWithScoreOut] = None

    @classmethod
    def from_instance(
        cls,
        score_bet: "ScoreBetModel",
        *,
        locked: bool,
        lang: Lang,
    ) -> "ScoreBetOut":
        return cls(
            id=score_bet.id,
            locked=locked,
            team1=(
                TeamWithScoreOut(
                    id=score_bet.match.team1.id,
                    code=score_bet.match.team1.code,
                    description=get_language_description(score_bet.match.team1, lang),
                    score=score_bet.score1,
                    flag=FlagOut(url=score_bet.match.team1.flag_url),
                )
                if score_bet.match.team1 is not None
                else None
            ),
            team2=(
                TeamWithScoreOut(
                    id=score_bet.match.team2.id,
                    code=score_bet.match.team2.code,
                    description=get_language_description(score_bet.match.team2, lang),
                    score=score_bet.score2,
                    flag=FlagOut(url=score_bet.match.team2.flag_url),
                )
                if score_bet.match.team2 is not None
                else None
            ),
        )


class Group(BaseModel):
    id: UUID4


class ScoreBetWithGroupIdOut(BaseModel):
    id: UUID4
    locked: bool
    group: Group
    team1: TeamWithScoreOut
    team2: TeamWithScoreOut

    @classmethod
    def from_instance(
        cls,
        score_bet: "ScoreBetModel",
        *,
        locked: bool,
        lang: Lang,
    ) -> "ScoreBetWithGroupIdOut":
        return cls(
            id=score_bet.id,
            locked=locked,
            group=Group(id=score_bet.match.group_id),
            team1=(
                TeamWithScoreOut(
                    id=score_bet.match.team1.id,
                    code=score_bet.match.team1.code,
                    description=get_language_description(score_bet.match.team1, lang),
                    score=score_bet.score1,
                    flag=FlagOut(url=score_bet.match.team1.flag_url),
                )
                if score_bet.match.team1
                else None
            ),
            team2=(
                TeamWithScoreOut(
                    id=score_bet.match.team2.id,
                    code=score_bet.match.team2.code,
                    description=get_language_description(score_bet.match.team2, lang),
                    score=score_bet.score2,
                    flag=FlagOut(url=score_bet.match.team2.flag_url),
                )
                if score_bet.match.team2
                else None
            ),
        )


class ScoreBetResponse(BaseModel):
    phase: PhaseOut
    group: GroupOut
    score_bet: ScoreBetOut


class ModifyScoreBetIn(BaseModel):
    team1: Optional[TeamModifyScoreBetIn] = None
    team2: Optional[TeamModifyScoreBetIn] = None

    model_config = ConfigDict(extra="forbid")



from copy import copy
from http import HTTPStatus
from secrets import SystemRandom, randbelow
from typing import TYPE_CHECKING, Any

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_group_rank_and_modify_score_bet(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_modify_bet_v2"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    # Signup user
    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(6),
            "first_name": get_random_string(10),
            "last_name": get_random_string(10),
            "password": get_random_string(12),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    authentication_token = response_signup.json()["result"]["token"]

    # Retrieve all the bets
    response_all_bets = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_all_bets.status_code == HTTPStatus.OK

    # Perform PATCH bets
    bet_ids = [score_bet["id"] for score_bet in response_all_bets.json()["result"]["score_bets"]]
    SystemRandom().shuffle(bet_ids)

    for bet_id in bet_ids:
        score1 = randbelow(5)
        score2 = randbelow(5)

        response = client.patch(
            f"/api/v1/score_bets/{bet_id}",
            json={"team1": {"score": score1}, "team2": {"score": score2}},
            headers={"Authorization": f"Bearer {authentication_token}"},
            follow_redirects=True,
        )
        assert response.status_code == HTTPStatus.OK
        assert response.json()["result"]["score_bet"]["id"] == bet_id
        assert response.json()["result"]["score_bet"]["team1"]["score"] == score1
        assert response.json()["result"]["score_bet"]["team2"]["score"] == score2

    # Retrieve group rank
    response_group_rank = client.get(
        "/api/v1/bets/groups/rank/A",
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_group_rank.status_code == HTTPStatus.OK

    group_rank = response_group_rank.json()["result"]["group_rank"]

    response_group_by_bet = client.get(
        "/api/v1/bets/groups/A",
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    score_bets = response_group_by_bet.json()["result"]["score_bets"]

    teams: dict[str, Any] = {}

    default_config = {
        "won": 0,
        "drawn": 0,
        "lost": 0,
        "goals_for": 0,
        "goals_against": 0,
        "points": 0,
    }

    for score_bet in score_bets:
        team1_id = score_bet["team1"]["id"]
        team2_id = score_bet["team2"]["id"]

        if team1_id not in teams:
            teams[team1_id] = copy(default_config)
            teams[team1_id]["team"] = {
                key: value for key, value in score_bet["team1"].items() if key != "score"
            }

        if team2_id not in teams:
            teams[team2_id] = copy(default_config)
            teams[team2_id]["team"] = {
                key: value for key, value in score_bet["team2"].items() if key != "score"
            }

        teams[team1_id]["goals_for"] += score_bet["team1"]["score"]
        teams[team1_id]["goals_against"] += score_bet["team2"]["score"]
        teams[team2_id]["goals_for"] += score_bet["team2"]["score"]
        teams[team2_id]["goals_against"] += score_bet["team1"]["score"]

        if score_bet["team1"]["score"] > score_bet["team2"]["score"]:
            teams[team1_id]["won"] += 1
            teams[team2_id]["lost"] += 1
        elif score_bet["team1"]["score"] == score_bet["team2"]["score"]:
            teams[team1_id]["drawn"] += 1
            teams[team2_id]["drawn"] += 1
        else:
            teams[team1_id]["lost"] += 1
            teams[team2_id]["won"] += 1

    for team in teams.values():
        team["played"] = team["won"] + team["drawn"] + team["lost"]
        team["goals_difference"] = team["goals_for"] - team["goals_against"]
        team["points"] = 3 * team["won"] + team["drawn"]

    assert sorted(group_rank, key=lambda group_position: group_position["team"]["id"]) == sorted(
        teams.values(),
        key=lambda group_position: group_position["team"]["id"],
    )



from http import HTTPStatus
from typing import TYPE_CHECKING
from unittest.mock import ANY

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_group(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_matches_db"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    # Signup one random user
    user_name = get_random_string(6)
    first_name = get_random_string(10)
    last_name = get_random_string(8)
    password = get_random_string(100)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": first_name,
            "last_name": last_name,
            "password": password,
        },
    )

    auth_token = response_signup.json()["result"]["token"]

    # Check all GET matches response
    expected_group = {"id": ANY, "code": "A", "phase": {"id": ANY}, "description": "Groupe A"}

    expected_group_without_phase = {
        key: value for key, value in expected_group.items() if key != "phase"
    }

    expected_phase = {
        "code": "GROUP",
        "description": "Phase de groupes",
        "id": ANY,
    }

    # Check groups associated to one phase
    group_response = client.get(
        "/api/v1/groups/phases/GROUP",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert group_response.status_code == HTTPStatus.OK
    assert group_response.json()["result"] == {
        "phase": expected_phase,
        "groups": [expected_group_without_phase],
    }

    # Check groups by phase with invalid phase identifier
    invalid_phase_code = get_random_string(5)

    group_response_invalid_phase_code = client.get(
        f"/api/v1/groups/phases/{invalid_phase_code}",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert group_response_invalid_phase_code.status_code == HTTPStatus.NOT_FOUND
    assert group_response_invalid_phase_code.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Phase not found: {invalid_phase_code}",
    }

    # Check GET /groups/{group_code}
    one_group_response = client.get(
        "/api/v1/groups/A",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert one_group_response.status_code == HTTPStatus.OK
    assert one_group_response.json()["result"] == {
        "group": expected_group_without_phase,
        "phase": expected_phase,
    }

    # Check GET /groups/{groupCode} with invalid code
    invalid_group_code = "B"

    group_response_with_invalid_code = client.get(
        "/api/v1/groups/B",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert group_response_with_invalid_code.status_code == HTTPStatus.NOT_FOUND
    assert group_response_with_invalid_code.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Group not found: {invalid_group_code}",
    }

    # Check GET /groups
    all_groups_response = client.get(
        "/api/v1/groups",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert all_groups_response.status_code == HTTPStatus.OK
    assert all_groups_response.json()["result"] == {
        "phases": [expected_phase],
        "groups": [expected_group],
    }



from typing import TYPE_CHECKING
from uuid import uuid4

from starlette.testclient import TestClient

from testing.util import get_random_string

if TYPE_CHECKING:
    from fastapi import FastAPI

QUERY_SIGNUP = """
    mutation Root(
        $userName: String!, $firstName: String!,
        $lastName: String!, $password: String!
    ) {
        signupResult(
            userName: $userName, firstName: $firstName,
            lastName: $lastName, password: $password
        ) {
            __typename
            ... on UserWithToken {
                id
                token
            }
            ... on UserNameAlreadyExists {
                message
            }
        }
    }
"""

QUERY_LOGIN = """
    mutation Root($userName: String!, $password: String!) {
        loginResult(userName: $userName, password: $password) {
            __typename
            ... on UserWithToken {
                firstName
                lastName
                fullName
            }
            ... on InvalidCredentials {
                message
            }
        }
    }
"""

QUERY_MODIFY_USER = """
    mutation Root($id: UUID!, $password: String!) {
        modifyUserResult(id: $id, password: $password) {
            __typename
            ... on UserWithoutSensitiveInfo {
                firstName
                lastName
                fullName
            }
            ... on UserNotFound {
                message
            }
            ... on InvalidToken {
                message
            }
            ... on ExpiredToken {
                message
            }
            ... on UnauthorizedAccessToAdminAPI {
                message
            }
        }
    }
"""


def test_modify_password(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    # Create admin account
    admin_name = "admin"

    response_signup_admin = client.post(
        "/api/v2",
        json={
            "query": QUERY_SIGNUP,
            "variables": {
                "userName": admin_name,
                "firstName": "admin",
                "lastName": "admin",
                "password": get_random_string(15),
            },
        },
    )

    assert response_signup_admin.json()["data"]["signupResult"]["__typename"] == "UserWithToken"
    authentication_token = response_signup_admin.json()["data"]["signupResult"]["token"]

    # Create non admin user account
    other_user_name = get_random_string(6)
    other_user_first_name = get_random_string(10)
    other_user_last_name = get_random_string(12)
    other_user_password = get_random_string(9)

    response_signup_glepape = client.post(
        "/api/v2",
        json={
            "query": QUERY_SIGNUP,
            "variables": {
                "userName": other_user_name,
                "firstName": other_user_first_name,
                "lastName": other_user_last_name,
                "password": other_user_password,
            },
        },
    )

    assert response_signup_glepape.json()["data"]["signupResult"]["__typename"] == "UserWithToken"
    user_id = response_signup_glepape.json()["data"]["signupResult"]["id"]
    authentication_token_glepape = response_signup_glepape.json()["data"]["signupResult"]["token"]

    # Check update is properly process
    new_password_other_user = get_random_string(15)

    response_modify_password = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={
            "query": QUERY_MODIFY_USER,
            "variables": {"id": user_id, "password": new_password_other_user},
        },
    )

    assert (
        response_modify_password.json()["data"]["modifyUserResult"]["__typename"]
        == "UserWithoutSensitiveInfo"
    )
    assert (
        response_modify_password.json()["data"]["modifyUserResult"]["fullName"]
        == f"{other_user_first_name} {other_user_last_name}"
    )

    # Check login successful with the new password
    response_login_glepape = client.post(
        "/api/v2",
        json={
            "query": QUERY_LOGIN,
            "variables": {"userName": other_user_name, "password": new_password_other_user},
        },
    )

    assert response_login_glepape.json()["data"]["loginResult"] == {
        "__typename": "UserWithToken",
        "firstName": other_user_first_name,
        "fullName": f"{other_user_first_name} {other_user_last_name}",
        "lastName": other_user_last_name,
    }

    # Check glepape user cannot update any password
    response_modify_password_with_glepape_user = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token_glepape}"},
        json={
            "query": QUERY_MODIFY_USER,
            "variables": {"id": user_id, "password": "new_new_password"},
        },
    )

    assert response_modify_password_with_glepape_user.json()["data"]["modifyUserResult"] == {
        "__typename": "UnauthorizedAccessToAdminAPI",
        "message": "Unauthorized access to admin API",
    }

    # Check call is rejected if user_id is invalid
    invalid_user_id = uuid4()

    response_wrong_input = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={
            "query": QUERY_MODIFY_USER,
            "variables": {"id": str(invalid_user_id), "password": "new_password_for_unknown_user"},
        },
    )

    assert response_wrong_input.json()["data"]["modifyUserResult"] == {
        "__typename": "UserNotFound",
        "message": f"User not found: {invalid_user_id}",
    }

    # Error case : invalid token
    response_modify_password = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {get_random_string(250)}"},
        json={
            "query": QUERY_MODIFY_USER,
            "variables": {"id": user_id, "password": new_password_other_user},
        },
    )

    assert response_modify_password.json()["data"]["modifyUserResult"] == {
        "__typename": "InvalidToken",
        "message": "Invalid token, authentication required",
    }



from typing import TYPE_CHECKING
from unittest.mock import ANY

from starlette.testclient import TestClient

from testing.util import get_random_string

if TYPE_CHECKING:
    from fastapi import FastAPI

QUERY_SIGNUP = """
    mutation Root(
        $userName: String!, $firstName: String!,
        $lastName: String!, $password: String!
    ) {
        signupResult(
            userName: $userName, firstName: $firstName,
            lastName: $lastName, password: $password
        ) {
            __typename
            ... on UserWithToken {
                firstName
                lastName
                fullName
                token
            }
            ... on UserNameAlreadyExists {
                message
            }
            ... on UnsatisfiedPasswordRequirements {
                message
            }
        }
    }
"""

QUERY_LOGIN = """
    mutation Root($userName: String!, $password: String!) {
        loginResult(userName: $userName, password: $password) {
            __typename
            ... on UserWithToken {
                token
            }
            ... on InvalidCredentials {
                message
            }
        }
    }
"""

QUERY_CURRENT_USER = """
    query {
        currentUserResult {
            __typename
            ... on User {
                fullName
            }
            ... on InvalidToken {
                message
            }
            ... on ExpiredToken {
                message
            }
        }
    }
"""


def test_signup_and_login(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    user_name = get_random_string(10)
    first_name = get_random_string(5)
    last_name = get_random_string(8)
    password = get_random_string(15)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": QUERY_SIGNUP,
            "variables": {
                "userName": user_name,
                "firstName": first_name,
                "lastName": last_name,
                "password": password,
            },
        },
    )

    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"

    response_login = client.post(
        "/api/v2",
        json={
            "query": QUERY_LOGIN,
            "variables": {
                "userName": user_name,
                "password": password,
            },
        },
    )

    assert response_login.json()["data"]["loginResult"]["__typename"] == "UserWithToken"

    auth_token = response_login.json()["data"]["loginResult"]["token"]

    response_current_user = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {auth_token}"},
        json={
            "query": QUERY_CURRENT_USER,
        },
    )

    assert response_current_user.json()["data"]["currentUserResult"]["__typename"] == "User"

    response_login_invalid_credentials = client.post(
        "/api/v2",
        json={
            "query": QUERY_LOGIN,
            "variables": {
                "userName": user_name,
                "password": get_random_string(6),
            },
        },
    )

    assert response_login_invalid_credentials.json() == {
        "data": {
            "loginResult": {"__typename": "InvalidCredentials", "message": "Invalid credentials"},
        },
    }


def test_signup_and_invalid_token(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    user_name = get_random_string(10)
    first_name = get_random_string(5)
    last_name = get_random_string(8)
    password = get_random_string(15)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": QUERY_SIGNUP,
            "variables": {
                "userName": user_name,
                "firstName": first_name,
                "lastName": last_name,
                "password": password,
            },
        },
    )
    assert response_signup.json()["data"]["signupResult"] == {
        "__typename": "UserWithToken",
        "firstName": first_name,
        "lastName": last_name,
        "fullName": f"{first_name} {last_name}",
        "token": ANY,
    }

    authentication_token = response_signup.json()["data"]["signupResult"]["token"]

    query_current_user = {
        "query": """
            query {
                currentUserResult {
                    __typename
                    ... on User {
                        firstName
                        lastName
                        fullName
                        result {
                            numberMatchGuess
                            points
                        }
                        groups {
                            description
                        }
                        phases {
                            description
                        }
                        scoreBets {
                            group {
                                id
                                description
                                phase {
                                    description
                                }
                            }
                            id
                            team1 {
                                score
                                description
                            }
                            team2 {
                                score
                                description
                            }
                        }
                    }
                    ... on InvalidToken {
                        message
                    }
                    ... on ExpiredToken {
                        message
                    }
                }
            }
        """,
    }

    response_current_user = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json=query_current_user,
    )

    assert response_current_user.json()["data"]["currentUserResult"]["__typename"] == "User"
    assert response_current_user.json()["data"]["currentUserResult"]["groups"] == []
    assert response_current_user.json()["data"]["currentUserResult"]["phases"] == []

    # invalidate authentication token and check currentUser query
    # send InvalidToken response
    response_current_user = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token[:-1]}"},
        json=query_current_user,
    )

    assert response_current_user.json()["data"]["currentUserResult"] == {
        "__typename": "InvalidToken",
        "message": "Invalid token, authentication required",
    }

    response_current_user_invalid_key = client.post(
        "/api/v2",
        headers={"Authorization": f"InvalidKey {authentication_token[:-1]}"},
        json=query_current_user,
    )

    assert response_current_user_invalid_key.json()["data"]["currentUserResult"] == {
        "__typename": "InvalidToken",
        "message": "Invalid token, authentication required",
    }


def test_name_already_exists(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    user_name = get_random_string(8)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": QUERY_SIGNUP,
            "variables": {
                "userName": user_name,
                "firstName": get_random_string(8),
                "lastName": f"{get_random_string(4)} {get_random_string(5)}",
                "password": get_random_string(10),
            },
        },
    )
    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"

    response_signup_2 = client.post(
        "/api/v2",
        json={
            "query": QUERY_SIGNUP,
            "variables": {
                "userName": user_name,
                "firstName": get_random_string(7),
                "lastName": get_random_string(10),
                "password": get_random_string(5),
            },
        },
    )
    assert response_signup_2.json()["data"]["signupResult"] == {
        "__typename": "UserNameAlreadyExists",
        "message": f"Name already exists: {user_name}",
    }


def test_expired_token(app_with_null_jwt_expiration_time: "FastAPI") -> None:
    client = TestClient(app_with_null_jwt_expiration_time)

    user_name = get_random_string(6)
    password = get_random_string(15)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": QUERY_SIGNUP,
            "variables": {
                "userName": user_name,
                "firstName": get_random_string(2),
                "lastName": get_random_string(8),
                "password": password,
            },
        },
    )

    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"

    auth_token = response_signup.json()["data"]["signupResult"]["token"]

    response_expired_token = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {auth_token}"},
        json={
            "query": QUERY_CURRENT_USER,
        },
    )

    assert response_expired_token.json() == {
        "data": {
            "currentUserResult": {
                "__typename": "ExpiredToken",
                "message": "Expired token, re-authentication required",
            },
        },
    }


def test_non_compliant_password(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": QUERY_SIGNUP,
            "variables": {
                "userName": get_random_string(3),
                "firstName": get_random_string(2),
                "lastName": get_random_string(8),
                "password": get_random_string(6),
            },
        },
    )

    assert response_signup.json() == {
        "data": {
            "signupResult": {
                "__typename": "UnsatisfiedPasswordRequirements",
                "message": "Password is too short. Minimum size is 8.",
            }
        }
    }



from http import HTTPStatus
from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_phase(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_phase_v1"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    # Signup one random user
    user_name = get_random_string(6)
    first_name = get_random_string(10)
    last_name = get_random_string(8)
    password = get_random_string(21)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": first_name,
            "last_name": last_name,
            "password": password,
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    token = response_signup.json()["result"]["token"]

    # Success case : retrieve all phases
    response_all_phases = client.get("/api/v1/phases", headers={"Authorization": f"Bearer {token}"})

    assert response_all_phases.status_code == HTTPStatus.OK
    assert response_all_phases.json() == {
        "ok": True,
        "result": [
            {"code": "GROUP", "description": "Phase de groupes", "id": ANY},
            {"code": "FINAL", "description": "Phase finale", "id": ANY},
        ],
    }

    # Success case : retrieve phase by id
    phase_id = response_all_phases.json()["result"][0]["id"]

    response_phase_by_id = client.get(
        f"/api/v1/phases/{phase_id}",
        headers={"Authorization": f"Bearer {token}"},
    )

    assert response_phase_by_id.status_code == HTTPStatus.OK
    assert response_phase_by_id.json() == {
        "ok": True,
        "result": {"code": "GROUP", "description": "Phase de groupes", "id": phase_id},
    }

    # Error case : retrieve phase by invalid id
    invalid_phase_id = uuid4()

    response_phase_with_invalid_id = client.get(
        f"/api/v1/phases/{invalid_phase_id}",
        headers={"Authorization": f"Bearer {token}"},
    )

    assert response_phase_with_invalid_id.status_code == HTTPStatus.NOT_FOUND
    assert response_phase_with_invalid_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Phase not found: {invalid_phase_id}",
    }



import re
from dataclasses import dataclass


@dataclass(init=False, eq=False, frozen=True)
class PasswordRequirements:
    MINIMUM_LENGTH: int = 8
    UPPERCASE: bool = True
    LOWERCASE: bool = True
    DIGIT: bool = True
    NO_SPACE: bool = True


class PasswordRequirementsError(Exception):
    pass


class TooShortError(PasswordRequirementsError):
    def __init__(self, minimum_length: int) -> None:
        super().__init__(f"Password is too short. Minimum size is {minimum_length}.")


class NoUpperCaseError(PasswordRequirementsError):
    def __init__(self) -> None:
        super().__init__("At least one upper-case letter expected.")


class NoLowerCaseError(PasswordRequirementsError):
    def __init__(self) -> None:
        super().__init__("At least one lower-case letter expected.")


class NoDigitError(PasswordRequirementsError):
    def __init__(self) -> None:
        super().__init__("At least one digit expected.")


class SpaceError(PasswordRequirementsError):
    def __init__(self) -> None:
        super().__init__("Password must not contain spaces.")


def validate_password(password: str) -> None:
    password_requirements = PasswordRequirements()

    if len(password) < password_requirements.MINIMUM_LENGTH:
        raise TooShortError(password_requirements.MINIMUM_LENGTH)

    if not re.search(r"[A-Z]", password) and password_requirements.UPPERCASE:
        raise NoUpperCaseError

    if not re.search(r"[a-z]", password) and password_requirements.LOWERCASE:
        raise NoLowerCaseError

    if not re.search(r"[0-9]", password) and password_requirements.DIGIT:
        raise NoDigitError

    if " " in password and password_requirements.NO_SPACE:
        raise SpaceError



from typing import TYPE_CHECKING

import pytest

from yak_server.cli.database.sync import (
    SyncOfficialResultsNotAvailableError,
    synchronize_official_results,
)

if TYPE_CHECKING:
    from sqlalchemy import Engine


def test_db_sync(monkeypatch: "pytest.MonkeyPatch", engine_for_test: "Engine") -> None:
    monkeypatch.setattr("yak_server.cli.database.sync.bs4", None)

    with pytest.raises(SyncOfficialResultsNotAvailableError) as exception:
        synchronize_official_results(engine_for_test)

    assert str(exception.value) == (
        "Synchronize official results is not available without sync extra dependency installed. "
        "To enable it, please run: uv pip install yak-server[sync]"
    )



from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_bets(
    app_with_lock_datetime_in_past: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    client = TestClient(app_with_lock_datetime_in_past)

    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_modify_bet_v2"),
    )

    initialize_database(engine_for_test, app_with_lock_datetime_in_past)

    user_name = get_random_string(10)
    first_name = get_random_string(5)
    last_name = get_random_string(8)
    password = get_random_string(15)

    query_signup = """
        mutation Root(
            $userName: String!, $firstName: String!,
            $lastName: String!, $password: String!
        ) {
            signupResult(
                userName: $userName, firstName: $firstName,
                lastName: $lastName, password: $password
            ) {
                __typename
                ... on UserWithToken {
                    token
                    scoreBets {
                        id
                    }
                }
                ... on UserNameAlreadyExists {
                    message
                }
            }
        }
    """

    response_signup = client.post(
        "/api/v2",
        json={
            "query": query_signup,
            "variables": {
                "userName": user_name,
                "firstName": first_name,
                "lastName": last_name,
                "password": password,
            },
        },
    )

    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"
    authentication_token = response_signup.json()["data"]["signupResult"]["token"]

    score_bet_ids = [
        bet["id"] for bet in response_signup.json()["data"]["signupResult"]["scoreBets"]
    ]

    query_score_bet = """
        query getScoreBet($scoreBetId: UUID!) {
            scoreBetResult(id: $scoreBetId) {
                __typename
                ... on ScoreBet {
                    id
                    locked
                    group {
                        id
                        description
                    }
                    team1 {
                        score
                        description
                    }
                    team2 {
                        score
                        description
                    }
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on ScoreBetNotFound {
                    message
                }
            }
        }
    """

    # Success case : Retrieve one score bet
    response_score_bet = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={"query": query_score_bet, "variables": {"scoreBetId": score_bet_ids[0]}},
    )

    assert response_score_bet.json() == {
        "data": {
            "scoreBetResult": {
                "__typename": "ScoreBet",
                "id": ANY,
                "locked": True,
                "group": {"description": "Groupe A", "id": ANY},
                "team1": {"description": "Croatie", "score": None},
                "team2": {"description": "Finlande", "score": None},
            },
        },
    }

    # Error case : Bet with invalid id
    invalid_bet_id = str(uuid4())

    response_score_bet_with_invalid_id = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={"query": query_score_bet, "variables": {"scoreBetId": invalid_bet_id}},
    )

    assert response_score_bet_with_invalid_id.json() == {
        "data": {
            "scoreBetResult": {
                "__typename": "ScoreBetNotFound",
                "message": f"Score bet not found: {invalid_bet_id}",
            },
        },
    }

    # Error case : authentication error
    response_authentication_error = client.post(
        "/api/v2",
        headers={"Authorization": "Bearer llllllllllllll"},
        json={"query": query_score_bet, "variables": {"scoreBetId": score_bet_ids[0]}},
    )

    assert response_authentication_error.json() == {
        "data": {
            "scoreBetResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }



import logging
from logging.config import fileConfig
from typing import TYPE_CHECKING

from alembic import context

from yak_server.database import build_engine
from yak_server.database.models import Base

if TYPE_CHECKING:
    from sqlalchemy.sql.schema import MetaData

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)
logger = logging.getLogger("alembic.env")


def get_engine_url() -> str:
    engine = build_engine()

    try:
        return engine.url.render_as_string(hide_password=False).replace("%", "%%")
    except AttributeError:
        return str(engine.url).replace("%", "%%")


# add your model's MetaData object here
# for 'autogenerate' support
config.set_main_option("sqlalchemy.url", get_engine_url())

# other values from the config, defined by the needs of env.py,
# can be acquired:
# ... etc.


def get_metadata() -> "MetaData":
    return Base.metadata


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(url=url, target_metadata=get_metadata(), literal_binds=True)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    engine = build_engine()

    with engine.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=get_metadata(),
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from http import HTTPStatus
from typing import TYPE_CHECKING
from uuid import uuid4

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_modify_partial_score_bet(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_modify_bet_v2"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(10),
            "first_name": get_random_string(5),
            "last_name": get_random_string(8),
            "password": get_random_string(15),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED
    authentication_token = response_signup.json()["result"]["token"]

    response_get_all_bets = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_get_all_bets.status_code == HTTPStatus.OK

    score_bet_ids = [
        score_bet["id"] for score_bet in response_get_all_bets.json()["result"]["score_bets"]
    ]

    # Success case : modify teams and scores
    score1 = 0
    score2 = 4

    new_team1 = response_get_all_bets.json()["result"]["score_bets"][-1]["team1"]
    new_team2 = response_get_all_bets.json()["result"]["score_bets"][-2]["team1"]

    response_patch_bet = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        json={
            "team1": {"id": new_team1["id"], "score": score1},
            "team2": {"id": new_team2["id"], "score": score2},
        },
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_patch_bet.status_code == HTTPStatus.OK
    assert response_patch_bet.json()["result"]["score_bet"]["team1"] == {
        **new_team1,
        "score": score1,
    }
    assert response_patch_bet.json()["result"]["score_bet"]["team2"] == {
        **new_team2,
        "score": score2,
    }

    # Success case : Patch only the team1 id
    new_team1 = response_get_all_bets.json()["result"]["score_bets"][-3]["team1"]

    response_patch_only_team1_id = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        json={"team1": {"id": new_team1["id"]}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_patch_only_team1_id.status_code == HTTPStatus.OK
    assert response_patch_only_team1_id.json()["result"]["score_bet"]["team1"] == {
        **new_team1,
        "score": score1,
    }

    # Success case : Patch only the team2 id
    new_team2 = response_get_all_bets.json()["result"]["score_bets"][-3]["team2"]

    response_patch_only_team2_id = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        json={"team2": {"id": new_team2["id"]}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_patch_only_team2_id.status_code == HTTPStatus.OK
    assert response_patch_only_team2_id.json()["result"]["score_bet"]["team2"] == {
        **new_team2,
        "score": score2,
    }

    # Success case : no modification
    response_no_modification = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        json={},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_no_modification.status_code == HTTPStatus.OK
    assert response_no_modification.json() == response_patch_only_team2_id.json()

    # Error case : Patch team1 id with wrong non existing id
    wrong_team1_id = str(uuid4())

    response_patch_wrong_team1_id = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        json={"team1": {"id": wrong_team1_id}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_patch_wrong_team1_id.status_code == HTTPStatus.NOT_FOUND
    assert response_patch_wrong_team1_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Team not found: {wrong_team1_id}",
    }

    # Error case : Patch team2 id with wrong non existing id
    wrong_team2_id = str(uuid4())

    response_patch_wrong_team2_id = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        json={"team2": {"id": wrong_team2_id}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_patch_wrong_team2_id.status_code == HTTPStatus.NOT_FOUND
    assert response_patch_wrong_team2_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Team not found: {wrong_team2_id}",
    }



from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_group_rank(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    client = TestClient(app_with_valid_jwt_config)

    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_compute_points_v1"),
    )
    initialize_database(engine_for_test, app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": """
                mutation Root(
                    $userName: String!, $firstName: String!,
                    $lastName: String!, $password: String!
                ) {
                    signupResult(
                        userName: $userName, firstName: $firstName,
                        lastName: $lastName, password: $password
                    ) {
                        __typename
                        ... on UserWithToken {
                            token
                            scoreBets {
                                id
                            }
                        }
                        ... on UserNameAlreadyExists {
                            message
                        }
                    }
                }
            """,
            "variables": {
                "userName": get_random_string(6),
                "firstName": get_random_string(10),
                "lastName": get_random_string(10),
                "password": get_random_string(10),
            },
        },
    )

    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"

    token = response_signup.json()["data"]["signupResult"]["token"]

    query_modify_score_bet = """
        mutation Root($id: UUID!, $score1: Int!, $score2: Int!) {
            modifyScoreBetResult(id: $id, score1: $score1, score2: $score2) {
                __typename
                ... on ScoreBet {
                    id
                }
                ... on ScoreBetNotFoundForUpdate {
                    message
                }
                ... on LockedScoreBetError {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on InvalidToken {
                    message
                }
            }
        }
    """

    # Modify score bet
    new_scores = [(5, 1), (0, 0), (1, 2)]

    for bet, new_score in zip(
        response_signup.json()["data"]["signupResult"]["scoreBets"],
        new_scores,
    ):
        response_patch_bet = client.post(
            "/api/v2",
            headers={"Authorization": f"Bearer {token}"},
            json={
                "query": query_modify_score_bet,
                "variables": {
                    "id": bet["id"],
                    "score1": new_score[0],
                    "score2": new_score[1],
                },
            },
        )

        assert response_patch_bet.json()["data"]["modifyScoreBetResult"]["__typename"] == "ScoreBet"

    query_group_rank = """
        query Root($code: String!) {
            groupRankByCodeResult(code: $code) {
                __typename
                ... on GroupRank {
                    group {
                        id
                        description
                    }
                    groupRank {
                        team {
                            description
                            code
                            id
                            flag {
                                url
                            }
                        }
                        played
                        won
                        drawn
                        lost
                        goalsFor
                        goalsAgainst
                        goalsDifference
                        points
                    }
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on GroupByCodeNotFound {
                    message
                }
            }
        }
    """

    # Success case : Fetch group rank and check updates
    response_group_rank_by_code = client.post(
        "/api/v2",
        json={"query": query_group_rank, "variables": {"code": "A"}},
        headers={"Authorization": f"Bearer {token}"},
    )

    result_group_rank = {
        "__typename": "GroupRank",
        "group": {"id": ANY, "description": "Groupe A"},
        "groupRank": [
            {
                "drawn": 1,
                "goalsAgainst": 1,
                "goalsDifference": 4,
                "goalsFor": 5,
                "lost": 0,
                "played": 2,
                "points": 4,
                "team": {
                    "description": "France",
                    "code": "FR",
                    "id": ANY,
                    "flag": {"url": ANY},
                },
                "won": 1,
            },
            {
                "drawn": 1,
                "goalsAgainst": 1,
                "goalsDifference": 1,
                "goalsFor": 2,
                "lost": 0,
                "played": 2,
                "points": 4,
                "team": {
                    "description": "Île de Man",
                    "code": "IM",
                    "id": ANY,
                    "flag": {"url": ANY},
                },
                "won": 1,
            },
            {
                "drawn": 0,
                "goalsAgainst": 7,
                "goalsDifference": -5,
                "goalsFor": 2,
                "lost": 2,
                "played": 2,
                "points": 0,
                "team": {
                    "description": "Irlande",
                    "code": "IE",
                    "id": ANY,
                    "flag": {"url": ANY},
                },
                "won": 0,
            },
        ],
    }

    assert response_group_rank_by_code.json()["data"] == {
        "groupRankByCodeResult": result_group_rank,
    }

    # Error case : Retrieve group rank with invalid group code
    group_id = response_group_rank_by_code.json()["data"]["groupRankByCodeResult"]["group"]["id"]

    invalid_group_code = "B"

    response_group_rank_with_invalid_code = client.post(
        "/api/v2",
        json={"query": query_group_rank, "variables": {"code": invalid_group_code}},
        headers={"Authorization": f"Bearer {token}"},
    )

    assert response_group_rank_with_invalid_code.json()["data"] == {
        "groupRankByCodeResult": {
            "__typename": "GroupByCodeNotFound",
            "message": f"Group not found: {invalid_group_code}",
        },
    }

    # Error case : Retrieve group rank with invalid token
    response_group_rank_with_auth_error = client.post(
        "/api/v2",
        json={"query": query_group_rank, "variables": {"code": "A"}},
        headers={"Authorization": f"Bearer {token[:-1]}"},
    )

    assert response_group_rank_with_auth_error.json()["data"] == {
        "groupRankByCodeResult": {
            "__typename": "InvalidToken",
            "message": "Invalid token, authentication required",
        },
    }

    query_group_rank_by_id = """
        query Root($id: UUID!) {
            groupRankByIdResult(id: $id) {
                __typename
                ... on GroupRank {
                    group {
                        id
                        description
                    }
                    groupRank {
                        team {
                            description
                            code
                            id
                            flag {
                                url
                            }
                        }
                        played
                        won
                        drawn
                        lost
                        goalsFor
                        goalsAgainst
                        goalsDifference
                        points
                    }
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on GroupByIdNotFound {
                    message
                }
            }
        }
    """

    # Success case : retrieve group rank with group id
    response_group_rank_by_id = client.post(
        "/api/v2",
        json={"query": query_group_rank_by_id, "variables": {"id": group_id}},
        headers={"Authorization": f"Bearer {token}"},
    )

    assert response_group_rank_by_id.json()["data"] == {"groupRankByIdResult": result_group_rank}

    # Error case : invalid authentication
    response_group_rank_by_id_auth_error = client.post(
        "/api/v2",
        json={"query": query_group_rank_by_id, "variables": {"id": group_id}},
        headers={"Authorization": f"Bearer {token[:-1]}"},
    )

    assert response_group_rank_by_id_auth_error.json()["data"] == {
        "groupRankByIdResult": {
            "__typename": "InvalidToken",
            "message": "Invalid token, authentication required",
        },
    }

    # Error case : retrieve group rank with invalid id
    invalid_id = str(uuid4())

    response_group_rank_with_invalid_id = client.post(
        "/api/v2",
        json={"query": query_group_rank_by_id, "variables": {"id": invalid_id}},
        headers={"Authorization": f"Bearer {token}"},
    )

    assert response_group_rank_with_invalid_id.json()["data"] == {
        "groupRankByIdResult": {
            "__typename": "GroupByIdNotFound",
            "message": f"Group not found: {invalid_id}",
        },
    }

    # Modify score bet and check group rank by id updates
    response_modify_score_bet = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "query": query_modify_score_bet,
            "variables": {
                "id": bet["id"],
                "score1": 4,
                "score2": 3,
            },
        },
    )

    assert (
        response_modify_score_bet.json()["data"]["modifyScoreBetResult"]["__typename"] == "ScoreBet"
    )

    response_retrieve_group_rank_by_id = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "query": query_group_rank_by_id,
            "variables": {
                "id": group_id,
            },
        },
    )

    group_rank_result = {
        "__typename": "GroupRank",
        "group": {"description": "Groupe A", "id": group_id},
        "groupRank": [
            {
                "team": ANY,
                "played": 2,
                "won": 1,
                "drawn": 1,
                "lost": 0,
                "goalsFor": 5,
                "goalsAgainst": 1,
                "goalsDifference": 4,
                "points": 4,
            },
            {
                "team": ANY,
                "played": 2,
                "won": 1,
                "drawn": 0,
                "lost": 1,
                "goalsFor": 5,
                "goalsAgainst": 8,
                "goalsDifference": -3,
                "points": 3,
            },
            {
                "team": ANY,
                "played": 2,
                "won": 0,
                "drawn": 1,
                "lost": 1,
                "goalsFor": 3,
                "goalsAgainst": 4,
                "goalsDifference": -1,
                "points": 1,
            },
        ],
    }

    assert response_retrieve_group_rank_by_id.json()["data"] == {
        "groupRankByIdResult": group_rank_result,
    }

    # Success case : No recomputation for group rank
    response_retrieve_group_rank_by_code = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "query": query_group_rank,
            "variables": {
                "code": "A",
            },
        },
    )

    assert response_retrieve_group_rank_by_code.json()["data"] == {
        "groupRankByCodeResult": group_rank_result,
    }



from operator import itemgetter
from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_teams(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    client = TestClient(app_with_valid_jwt_config)

    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_teams_v1"),
    )
    initialize_database(engine_for_test, app_with_valid_jwt_config)

    user_name = get_random_string(6)
    first_name = get_random_string(10)
    last_name = get_random_string(2)
    password = get_random_string(20)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": """
                mutation Signup(
                    $userName: String!, $firstName: String!,
                    $lastName: String!, $password: String!
                ) {
                    signupResult(
                        userName: $userName, firstName: $firstName,
                        lastName: $lastName, password: $password
                    ) {
                        __typename
                        ... on UserWithToken {
                            token
                        }
                        ... on UserNameAlreadyExists {
                            message
                        }
                    }
                }
            """,
            "variables": {
                "userName": user_name,
                "firstName": first_name,
                "lastName": last_name,
                "password": password,
            },
        },
    )

    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"

    auth_token = response_signup.json()["data"]["signupResult"]["token"]

    all_teams_query = """
        query {
            allTeamsResult {
                __typename
                ... on AllTeamsSuccessful {
                    teams {
                        id
                        code
                        description
                        flag {
                            url
                        }
                    }
                }
                ... on InvalidToken {
                    __typename
                    message
                }
                ... on ExpiredToken {
                    __typename
                    message
                }
            }
        }
    """

    response_all_teams = client.post(
        "/api/v2",
        json={
            "query": all_teams_query,
        },
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert response_all_teams.json()["data"]["allTeamsResult"]["__typename"] == "AllTeamsSuccessful"
    expected_teams = [
        {
            "id": ANY,
            "code": "GR",
            "description": "Grèce",
            "flag": {"url": ANY},
        },
        {
            "id": ANY,
            "code": "DE",
            "description": "Allemagne",
            "flag": {"url": ANY},
        },
        {
            "id": ANY,
            "code": "GD",
            "description": "Grenade",
            "flag": {"url": ANY},
        },
        {
            "id": ANY,
            "code": "UA",
            "description": "Ukraine",
            "flag": {"url": ANY},
        },
        {
            "id": ANY,
            "code": "ML",
            "description": "Mali",
            "flag": {"url": ANY},
        },
        {
            "id": ANY,
            "code": "JM",
            "description": "Jamaïque",
            "flag": {"url": ANY},
        },
        {
            "id": ANY,
            "code": "JO",
            "description": "Jordanie",
            "flag": {"url": ANY},
        },
        {
            "id": ANY,
            "code": "NO",
            "description": "Norvège",
            "flag": {"url": ANY},
        },
    ]

    for team in expected_teams:
        assert f"/api/v1/teams/{team['id']}/flag" == team["flag"]["url"]

    assert sorted(
        response_all_teams.json()["data"]["allTeamsResult"]["teams"],
        key=itemgetter("code"),
    ) == sorted(expected_teams, key=itemgetter("code"))

    team_norway = [
        team
        for team in response_all_teams.json()["data"]["allTeamsResult"]["teams"]
        if team["description"] == "Norvège"
    ]

    assert len(team_norway) == 1

    team_norway_id = team_norway[0]["id"]

    response_all_teams_invalid_token = client.post(
        "/api/v2",
        json={
            "query": all_teams_query,
        },
        headers={"Authorization": f"Bearer {auth_token[:12]}{auth_token[15:]}"},
    )

    assert (
        response_all_teams_invalid_token.json()["data"]["allTeamsResult"]["__typename"]
        == "InvalidToken"
    )

    team_by_id_query = """
        query Root($teamId: UUID!) {
            teamByIdResult(id: $teamId) {
                __typename
                ... on Team {
                    id
                    code
                    description
                    flag {
                        url
                    }
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on TeamByIdNotFound {
                    message
                }
            }
        }
    """

    response_team_by_id = client.post(
        "/api/v2",
        json={
            "query": team_by_id_query,
            "variables": {
                "teamId": team_norway_id,
            },
        },
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert response_team_by_id.json() == {
        "data": {
            "teamByIdResult": {
                "__typename": "Team",
                "code": "NO",
                "description": "Norvège",
                "flag": {"url": f"/api/v1/teams/{team_norway_id}/flag"},
                "id": team_norway_id,
            },
        },
    }

    invalid_team_id = str(uuid4())

    response_team_with_invalid_id = client.post(
        "/api/v2",
        json={
            "query": team_by_id_query,
            "variables": {
                "teamId": invalid_team_id,
            },
        },
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert response_team_with_invalid_id.json() == {
        "data": {
            "teamByIdResult": {
                "__typename": "TeamByIdNotFound",
                "message": f"Team not found: {invalid_team_id}",
            },
        },
    }

    response_authentication_error = client.post(
        "/api/v2",
        json={
            "query": team_by_id_query,
            "variables": {
                "teamId": invalid_team_id,
            },
        },
        headers={"Authorization": f"Bearer {auth_token[:-1]}"},
    )

    assert response_authentication_error.json() == {
        "data": {
            "teamByIdResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }

    team_by_code_query = """
        query Root($teamCode: String!) {
            teamByCodeResult(code: $teamCode) {
                __typename
                ... on Team {
                    id
                    code
                    description
                    flag {
                        url
                    }
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on TeamByCodeNotFound {
                    message
                }
            }
        }
    """

    response_team_by_code = client.post(
        "/api/v2",
        json={
            "query": team_by_code_query,
            "variables": {
                "teamCode": "NO",
            },
        },
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert response_team_by_code.json() == {
        "data": {
            "teamByCodeResult": {
                "__typename": "Team",
                "code": "NO",
                "description": "Norvège",
                "flag": {"url": f"/api/v1/teams/{team_norway_id}/flag"},
                "id": team_norway_id,
            },
        },
    }

    response_team_with_invalid_code = client.post(
        "/api/v2",
        json={
            "query": team_by_code_query,
            "variables": {
                "teamCode": "NA",
            },
        },
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert response_team_with_invalid_code.json() == {
        "data": {
            "teamByCodeResult": {
                "__typename": "TeamByCodeNotFound",
                "message": "Team not found: NA",
            },
        },
    }

    response_authentication_error = client.post(
        "/api/v2",
        json={
            "query": team_by_code_query,
            "variables": {
                "teamCode": "NO",
            },
        },
    )

    assert response_authentication_error.json() == {
        "data": {
            "teamByCodeResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }



from collections.abc import Iterable
from typing import TYPE_CHECKING, Optional

from sqlalchemy import and_

from .models import BinaryBetModel, GroupModel, MatchModel, PhaseModel, ScoreBetModel

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from .models import UserModel


def bets_from_group_code(
    db: "Session",
    user: "UserModel",
    group_code: str,
) -> tuple[Optional[GroupModel], Iterable[ScoreBetModel], Iterable[BinaryBetModel]]:
    group = db.query(GroupModel).filter_by(code=group_code).first()

    if not group:
        return None, [], []

    score_bets = (
        db.query(ScoreBetModel)
        .join(ScoreBetModel.match)
        .filter(and_(MatchModel.user_id == user.id, MatchModel.group_id == group.id))
        .order_by(MatchModel.index)
    )

    binary_bets = (
        db.query(BinaryBetModel)
        .join(BinaryBetModel.match)
        .filter(and_(MatchModel.user_id == user.id, MatchModel.group_id == group.id))
        .order_by(MatchModel.index)
    )

    return group, score_bets, binary_bets


def bets_from_phase_code(
    db: "Session",
    user: "UserModel",
    phase_code: str,
) -> tuple[
    Optional[PhaseModel], Iterable[GroupModel], Iterable[ScoreBetModel], Iterable[BinaryBetModel]
]:
    phase = db.query(PhaseModel).filter_by(code=phase_code).first()

    if not phase:
        return None, [], [], []

    groups = db.query(GroupModel).filter_by(phase_id=phase.id).order_by(GroupModel.index)

    binary_bets = (
        db.query(BinaryBetModel)
        .join(BinaryBetModel.match)
        .join(MatchModel.group)
        .filter(and_(MatchModel.user_id == user.id, GroupModel.phase_id == phase.id))
        .order_by(GroupModel.index, MatchModel.index)
    )

    score_bets = (
        db.query(ScoreBetModel)
        .join(ScoreBetModel.match)
        .join(MatchModel.group)
        .filter(and_(MatchModel.user_id == user.id, GroupModel.phase_id == phase.id))
        .order_by(GroupModel.index, MatchModel.index)
    )

    return phase, groups, score_bets, binary_bets



from functools import cache

import psycopg2
from pydantic_settings import BaseSettings, SettingsConfigDict
from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import Session, sessionmaker


class PostgresSettings(BaseSettings):
    host: str
    user_name: str
    password: str
    port: int
    db: str

    model_config = SettingsConfigDict(
        env_file=".env.db",
        env_file_encoding="utf-8",
        env_prefix="postgres_",
    )


@cache
def get_postgres_settings() -> PostgresSettings:
    return PostgresSettings()


def compute_database_uri(
    client: str, host: str, user_name: str, password: str, port: int, db: str
) -> str:
    return f"postgresql+{client}://{user_name}:{password}@{host}:{port}/{db}"


def build_engine() -> Engine:
    postgres_settings = get_postgres_settings()

    database_url = compute_database_uri(
        psycopg2.__name__,
        postgres_settings.host,
        postgres_settings.user_name,
        postgres_settings.password,
        postgres_settings.port,
        postgres_settings.db,
    )

    return create_engine(database_url, pool_recycle=7200, pool_pre_ping=True)


def build_local_session_maker(engine: Engine) -> sessionmaker[Session]:
    return sessionmaker(autocommit=False, autoflush=False, bind=engine)



import json
import secrets
from enum import Enum
from pathlib import Path
from typing import Any
from uuid import UUID

import pendulum

from yak_server.database import PostgresSettings
from yak_server.helpers.rules import RULE_MAPPING, Rules


class YesOrNo(str, Enum):
    y = "y"
    n = "n"


class RuleNotDefinedError(Exception):
    def __init__(self, rule_id: UUID) -> None:
        super().__init__(f"Rule not defined: {rule_id}")


class InvalidLockDatetimeError(Exception):
    def __init__(self, raw_lock_datetime: str) -> None:
        super().__init__(f"lock_datetime is not a valid datetime: {raw_lock_datetime}")


def write_env_file(env: dict[str, Any], filename: str) -> None:
    Path(filename).write_text(
        "".join([f"{env_var}={env_value}\n" for env_var, env_value in env.items()]),
        encoding="utf-8",
    )


def write_app_env_file(debug: bool, jwt_expiration_time: int, competition: str) -> None:  # noqa: FBT001
    env: dict[str, Any] = {}

    env["DEBUG"] = 1 if debug else 0

    env["JWT_EXPIRATION_TIME"] = jwt_expiration_time
    env["JWT_SECRET_KEY"] = secrets.token_hex(128)

    # Select competition to load associated rules
    path = Path(__file__).parents[1] / "data"

    env["COMPETITION"] = competition

    data_folder = path / competition
    env["DATA_FOLDER"] = data_folder

    # Load rules in environment
    rules_list: dict[str, Any] = {}

    for rule_file in Path(data_folder, "rules").glob("*.json"):
        rule_id = UUID(rule_file.stem)

        if rule_id not in RULE_MAPPING:
            raise RuleNotDefinedError(rule_id)

        rule_name = RULE_MAPPING[rule_id].attribute

        rules_list[rule_name] = json.loads(rule_file.read_text())

    rules = Rules.model_validate(rules_list)
    env["RULES"] = rules.model_dump_json(exclude_unset=True)

    # Load lock datetime
    common_settings = json.loads((data_folder / "common.json").read_text())

    parsed_lock_datetime = pendulum.parse(common_settings["lock_datetime"], exact=True)

    if not isinstance(parsed_lock_datetime, pendulum.DateTime):
        raise InvalidLockDatetimeError(common_settings["lock_datetime"])

    env["LOCK_DATETIME"] = parsed_lock_datetime.to_iso8601_string()

    env["OFFICIAL_RESULTS_URL"] = common_settings["official_results_url"]

    write_env_file(env, ".env")


def write_db_env_file(host: str, user: str, password: str, port: int, db: str) -> None:
    # try to instantiate pydantic model to check if settings are ok
    PostgresSettings(host=host, user_name=user, password=password, port=port, db=db)

    env_db = {
        "POSTGRES_HOST": host,
        "POSTGRES_USER_NAME": user,
        "POSTGRES_PASSWORD": password,
        "POSTGRES_PORT": port,
        "POSTGRES_DB": db,
    }
    write_env_file(env_db, ".env.db")


def init_env(  # noqa: PLR0913, PLR0917
    debug: bool,  # noqa: FBT001
    host: str,
    db_username: str,
    password: str,
    competition: str,
    database: str,
    jwt_expiration: int,
    port: str,
) -> None:
    write_app_env_file(debug, jwt_expiration, competition)
    write_db_env_file(host, db_username, password, port, database)



import secrets
import string
from dataclasses import dataclass
from http import HTTPStatus
from pathlib import Path
from typing import Optional

from fastapi.testclient import TestClient

TESTING_PATH = Path(__file__).parent


def get_resources_path(path: str) -> Path:
    return TESTING_PATH / "resources" / path


def get_random_string(length: int) -> str:
    return (
        secrets.choice(string.digits)
        + secrets.choice(string.ascii_lowercase)
        + secrets.choice(string.ascii_uppercase)
        + "".join(secrets.choice(string.ascii_letters + string.digits) for _ in range(length - 3))
    )


@dataclass
class UserData:
    first_name: str
    last_name: str
    name: str
    scores: list[Optional[tuple[Optional[int], Optional[int]]]]
    token: str = ""


def patch_score_bets(
    client: TestClient,
    token: str,
    new_scores: list[Optional[tuple[Optional[int], Optional[int]]]],
) -> None:
    response_get_all_bets = client.get("/api/v1/bets", headers={"Authorization": f"Bearer {token}"})

    assert response_get_all_bets.status_code == HTTPStatus.OK

    for bet, new_score in zip(response_get_all_bets.json()["result"]["score_bets"], new_scores):
        if new_score is not None:
            response_patch_score_bet = client.patch(
                f"/api/v1/score_bets/{bet['id']}",
                headers={"Authorization": f"Bearer {token}"},
                json={
                    "team1": {"score": new_score[0]},
                    "team2": {"score": new_score[1]},
                },
            )

            assert response_patch_score_bet.status_code == HTTPStatus.OK



from typing import Annotated

from fastapi import Depends
from sqlalchemy.orm import Session
from strawberry.fastapi import BaseContext

from yak_server.helpers.database import get_db
from yak_server.helpers.settings import Settings, get_settings


class YakContext(BaseContext):
    db: Session
    settings: Settings

    def __init__(self, db: Session, settings: Settings) -> None:
        super().__init__()
        self.db = db
        self.settings = settings


def get_context(
    db: Annotated[Session, Depends(get_db)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> YakContext:
    return YakContext(db, settings)



import logging
from typing import Optional
from uuid import UUID

import pendulum
import strawberry
from sqlalchemy import and_
from strawberry.types import Info

from yak_server.database.models import (
    BinaryBetModel,
    MatchModel,
    ScoreBetModel,
    UserModel,
)
from yak_server.helpers.authentication import (
    NameAlreadyExistsError,
    encode_bearer_token,
    signup_user,
)
from yak_server.helpers.bet_locking import is_locked
from yak_server.helpers.group_position import set_recomputation_flag
from yak_server.helpers.logging_helpers import (
    logged_in_successfully,
    modify_binary_bet_successfully,
    modify_score_bet_successfully,
    signed_up_successfully,
)
from yak_server.helpers.password_validator import PasswordRequirementsError

from .bearer_authentication import authentify, authentify_admin
from .context import YakContext
from .result import (
    BinaryBetNotFoundForUpdate,
    InvalidCredentials,
    LockedBinaryBetError,
    LockedScoreBetError,
    LoginResult,
    ModifyBinaryBetResult,
    ModifyScoreBetResult,
    ModifyUserResult,
    NewScoreNegative,
    ScoreBetNotFoundForUpdate,
    SignupResult,
    UnsatisfiedPasswordRequirements,
    UserNameAlreadyExists,
    UserNotFound,
)
from .schema import BinaryBet, ScoreBet, UserWithoutSensitiveInfo, UserWithToken

logger = logging.getLogger(__name__)


@strawberry.type
class Mutation:
    @strawberry.mutation
    def signup_result(
        self,
        user_name: str,
        password: str,
        first_name: str,
        last_name: str,
        info: Info[YakContext, None],
    ) -> SignupResult:
        db = info.context.db
        settings = info.context.settings

        try:
            user = signup_user(
                db, name=user_name, first_name=first_name, last_name=last_name, password=password
            )
        except PasswordRequirementsError as password_requirements_error:
            return UnsatisfiedPasswordRequirements(message=str(password_requirements_error))
        except NameAlreadyExistsError:
            return UserNameAlreadyExists(user_name=user_name)

        logger.info(signed_up_successfully(user.name))

        return UserWithToken.from_instance_and_token(
            user,
            db=db,
            lock_datetime=settings.lock_datetime,
            token=encode_bearer_token(
                sub=user.id,
                expiration_time=pendulum.duration(seconds=settings.jwt_expiration_time),
                secret_key=settings.jwt_secret_key,
            ),
        )

    @strawberry.mutation
    def login_result(
        self,
        user_name: str,
        password: str,
        info: Info[YakContext, None],
    ) -> LoginResult:
        db = info.context.db
        settings = info.context.settings

        user = UserModel.authenticate(db=db, name=user_name, password=password)

        if not user:
            return InvalidCredentials()

        token = encode_bearer_token(
            sub=user.id,
            expiration_time=pendulum.duration(seconds=settings.jwt_expiration_time),
            secret_key=settings.jwt_secret_key,
        )

        logger.info(logged_in_successfully(user.name))

        return UserWithToken.from_instance_and_token(
            user,
            db=db,
            lock_datetime=settings.lock_datetime,
            token=token,
        )

    @strawberry.mutation
    def modify_binary_bet_result(
        self,
        id: UUID,
        is_one_won: Optional[bool],  # noqa: FBT001
        info: Info[YakContext, None],
    ) -> ModifyBinaryBetResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        bet = (
            db.query(BinaryBetModel)
            .join(BinaryBetModel.match)
            .filter(and_(MatchModel.user_id == user.id, BinaryBetModel.id == id))
            .first()
        )

        if is_locked(user.name, settings.lock_datetime):
            return LockedBinaryBetError()

        if not bet:
            return BinaryBetNotFoundForUpdate()

        logger.info(modify_binary_bet_successfully(user.name, bet, new_is_one_won=is_one_won))

        bet.is_one_won = is_one_won
        db.commit()

        return BinaryBet.from_instance(bet, db=db, lock_datetime=settings.lock_datetime)

    @strawberry.mutation
    def modify_score_bet_result(
        self,
        id: UUID,
        score1: Optional[int],
        score2: Optional[int],
        info: Info[YakContext, None],
    ) -> ModifyScoreBetResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        bet = (
            db.query(ScoreBetModel)
            .join(ScoreBetModel.match)
            .filter(and_(MatchModel.user_id == user.id, ScoreBetModel.id == id))
            .first()
        )

        if is_locked(user.name, settings.lock_datetime):
            return LockedScoreBetError()

        if not bet:
            return ScoreBetNotFoundForUpdate()

        if score1 is not None and score1 < 0:
            return NewScoreNegative(variable_name="$score1", score=score1)

        if score2 is not None and score2 < 0:
            return NewScoreNegative(variable_name="$score2", score=score2)

        if score1 == bet.score1 and score2 == bet.score2:
            return ScoreBet.from_instance(bet, db=db, lock_datetime=settings.lock_datetime)

        set_recomputation_flag(db, bet.match.team1_id, user.id)
        set_recomputation_flag(db, bet.match.team2_id, user.id)

        logger.info(modify_score_bet_successfully(user.name, bet, score1, score2))

        bet.score1 = score1
        bet.score2 = score2
        db.commit()

        return ScoreBet.from_instance(bet, db=db, lock_datetime=settings.lock_datetime)

    @strawberry.mutation
    def modify_user_result(
        self,
        id: UUID,
        password: str,
        info: Info[YakContext, None],
    ) -> ModifyUserResult:
        result = authentify_admin(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db

        user = db.query(UserModel).filter_by(id=id).first()

        if not user:
            return UserNotFound(id=id)

        user.change_password(password)
        db.commit()

        return UserWithoutSensitiveInfo.from_instance(user)



from collections.abc import Generator
from http import HTTPStatus
from typing import TYPE_CHECKING
from unittest.mock import ANY

import pendulum
import pytest
from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import UserData, get_random_string, patch_score_bets
from yak_server.cli.database import initialize_database
from yak_server.helpers.rules import Rules
from yak_server.helpers.rules.compute_final_from_rank import (
    RuleComputeFinaleFromGroupRank,
    Team,
    Versus,
)
from yak_server.helpers.rules.compute_points import RuleComputePoints
from yak_server.helpers.settings import get_settings

if TYPE_CHECKING:
    from fastapi import FastAPI
    from sqlalchemy import Engine


@pytest.fixture
def app_with_rules_and_score_board_config(
    app_with_valid_jwt_config: "FastAPI",
) -> Generator["FastAPI", None, None]:
    app_with_valid_jwt_config.dependency_overrides[get_settings] = MockSettings(
        jwt_expiration_time=100,
        jwt_secret_key=get_random_string(100),
        lock_datetime_shift=pendulum.duration(minutes=10),
        rules=Rules(
            compute_finale_phase_from_group_rank=RuleComputeFinaleFromGroupRank(
                to_group="2",
                from_phase="GROUP",
                versus=[
                    Versus(team1=Team(rank=1, group="A"), team2=Team(rank=2, group="B")),
                    Versus(team1=Team(rank=1, group="B"), team2=Team(rank=2, group="A")),
                ],
            ),
            compute_points=RuleComputePoints(
                base_correct_result=1,
                multiplying_factor_correct_result=2,
                base_correct_score=3,
                multiplying_factor_correct_score=7,
                team_qualified=10,
                first_team_qualified=20,
            ),
        ),
        base_correct_result=1,
        multiplying_factor_correct_result=2,
        base_correct_score=3,
        multiplying_factor_correct_score=7,
        team_qualified=10,
        first_team_qualified=20,
    )

    yield app_with_valid_jwt_config

    app_with_valid_jwt_config.dependency_overrides.clear()


def test_compute_points(
    app_with_rules_and_score_board_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    client = TestClient(app_with_rules_and_score_board_config)

    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_compute_points_edge_cases_v1"),
    )
    initialize_database(engine_for_test, app_with_rules_and_score_board_config)

    admin = UserData(
        name="admin",
        first_name="admin",
        last_name="admin",
        scores=[
            (1, 2),
            (5, 1),
            (5, 5),
            (2, 2),
            (3, 0),
            (2, 5),
            (5, 3),
            (1, 1),
            (6, 5),
            (1, 1),
            (1, 1),
            (2, 0),
        ],
    )

    response_signup_admin = client.post(
        "/api/v1/users/signup",
        json={
            "name": admin.name,
            "first_name": admin.first_name,
            "last_name": admin.last_name,
            "password": get_random_string(15),
        },
    )

    assert response_signup_admin.status_code == HTTPStatus.CREATED

    admin.token = response_signup_admin.json()["result"]["token"]

    # Patch score bet admin
    patch_score_bets(client, admin.token, admin.scores)

    # Signup 3 players and patch their scores
    users_data = [
        UserData(
            name=get_random_string(6),
            first_name=get_random_string(15),
            last_name=get_random_string(16),
            scores=[
                (4, 1),
                (2, 6),
                (1, 2),
                (4, 3),
                (1, 1),
                (0, 0),
                (2, 2),
                (3, 0),
                (6, 0),
                (None, 2),
                (None, 3),
                (5, 0),
            ],
        ),
        UserData(
            name=get_random_string(7),
            first_name=get_random_string(89),
            last_name=get_random_string(25),
            scores=[
                (4, 5),
                (3, 0),
                (None, 3),
                (0, 3),
                (0, None),
                (5, 4),
                (2, None),
                (1, 5),
                (0, 3),
                (None, None),
                (2, None),
                (2, None),
            ],
        ),
        UserData(
            name=get_random_string(26),
            first_name=get_random_string(14),
            last_name=get_random_string(85),
            scores=[
                (2, 6),
                (0, None),
                (0, None),
                (3, 2),
                (2, None),
                (0, 3),
                (6, 3),
                (0, 5),
                (3, 4),
                (2, None),
                (6, 4),
                (3, None),
            ],
        ),
    ]

    for user in users_data:
        response_signup = client.post(
            "/api/v1/users/signup",
            json={
                "first_name": user.first_name,
                "last_name": user.last_name,
                "name": user.name,
                "password": get_random_string(85),
            },
        )

        assert response_signup.status_code == HTTPStatus.CREATED

        user.token = response_signup.json()["result"]["token"]

        patch_score_bets(client, user.token, user.scores)

    # Success case : compute_points call
    response_compute_points = client.post(
        "/api/v1/rules/62d46542-8cf1-4a3b-af77-a5086f10ac59",
        headers={"Authorization": f"Bearer {admin.token}"},
    )

    assert response_compute_points.status_code == HTTPStatus.OK

    # Check final rule execution
    rule_final_phase_response = client.post(
        "/api/v1/rules/492345de-8d4a-45b6-8b94-d219f2b0c3e9",
        headers={"Authorization": f"Bearer {users_data[0].token}"},
    )

    assert rule_final_phase_response.status_code == HTTPStatus.OK

    # Check retrieve all bets after final phase
    bets_final_phase_response = client.get(
        "/api/v1/bets/phases/FINAL", headers={"Authorization": f"Bearer {users_data[0].token}"}
    )

    assert bets_final_phase_response.status_code == HTTPStatus.OK

    semi_finals_id = bets_final_phase_response.json()["result"]["groups"][0]["id"]
    finale_id = bets_final_phase_response.json()["result"]["groups"][1]["id"]

    assert bets_final_phase_response.json() == {
        "ok": True,
        "result": {
            "phase": {"id": ANY, "code": "FINAL", "description": "Phase finale"},
            "groups": [
                {"id": semi_finals_id, "code": "2", "description": "Demi-finale"},
                {"id": finale_id, "code": "1", "description": "Final"},
            ],
            "score_bets": [],
            "binary_bets": [
                {
                    "id": ANY,
                    "locked": False,
                    "group": {"id": semi_finals_id},
                    "team1": None,
                    "team2": None,
                },
                {
                    "id": ANY,
                    "locked": False,
                    "group": {"id": semi_finals_id},
                    "team1": None,
                    "team2": None,
                },
                {
                    "id": ANY,
                    "locked": False,
                    "group": {"id": finale_id},
                    "team1": None,
                    "team2": None,
                },
            ],
        },
    }

    # Success case : Check score board call
    score_board_response = client.get(
        "/api/v1/score_board",
        headers={"Authorization": f"Bearer {users_data[0].token}"},
    )

    assert score_board_response.json()["result"] == [
        {
            "rank": 1,
            "first_name": users_data[0].first_name,
            "last_name": users_data[0].last_name,
            "full_name": f"{users_data[0].first_name} {users_data[0].last_name}",
            "number_match_guess": 2,
            "number_score_guess": 0,
            "number_qualified_teams_guess": 1,
            "number_first_qualified_guess": 0,
            "number_quarter_final_guess": 0,
            "number_semi_final_guess": 0,
            "number_final_guess": 0,
            "number_winner_guess": 0,
            "points": 16.0,
        },
        {
            "rank": 2,
            "first_name": users_data[2].first_name,
            "last_name": users_data[2].last_name,
            "full_name": f"{users_data[2].first_name} {users_data[2].last_name}",
            "number_match_guess": 3,
            "number_score_guess": 0,
            "number_qualified_teams_guess": 0,
            "number_first_qualified_guess": 0,
            "number_quarter_final_guess": 0,
            "number_semi_final_guess": 0,
            "number_final_guess": 0,
            "number_winner_guess": 0,
            "points": 8.0,
        },
        {
            "rank": 3,
            "first_name": users_data[1].first_name,
            "last_name": users_data[1].last_name,
            "full_name": f"{users_data[1].first_name} {users_data[1].last_name}",
            "number_match_guess": 2,
            "number_score_guess": 0,
            "number_qualified_teams_guess": 0,
            "number_first_qualified_guess": 0,
            "number_quarter_final_guess": 0,
            "number_semi_final_guess": 0,
            "number_final_guess": 0,
            "number_winner_guess": 0,
            "points": 5.0,
        },
    ]

    # Patch admin score bets so all groups are not filled
    patch_score_bets(
        client,
        admin.token,
        [
            (1, 0),
            None,
            (1, 0),
            None,
            (None, 0),
            (5, 2),
            (None, None),
            (1, 0),
            (2, 0),
            (None, 5),
            (2, None),
            (None, None),
        ],
    )

    # Compute points again
    response_compute_points = client.post(
        "/api/v1/rules/62d46542-8cf1-4a3b-af77-a5086f10ac59",
        headers={"Authorization": f"Bearer {admin.token}"},
    )

    # Check again score board
    score_board_response = client.get(
        "/api/v1/score_board",
        headers={"Authorization": f"Bearer {users_data[0].token}"},
    )

    assert score_board_response.json()["result"] == [
        {
            "rank": 1,
            "first_name": users_data[0].first_name,
            "last_name": users_data[0].last_name,
            "full_name": f"{users_data[0].first_name} {users_data[0].last_name}",
            "number_match_guess": 3,
            "number_score_guess": 0,
            "number_qualified_teams_guess": 1,
            "number_first_qualified_guess": 0,
            "number_quarter_final_guess": 0,
            "number_semi_final_guess": 0,
            "number_final_guess": 0,
            "number_winner_guess": 0,
            "points": 9.0,
        },
        {
            "rank": 2,
            "first_name": users_data[1].first_name,
            "last_name": users_data[1].last_name,
            "full_name": f"{users_data[1].first_name} {users_data[1].last_name}",
            "number_match_guess": 2,
            "number_score_guess": 0,
            "number_qualified_teams_guess": 0,
            "number_first_qualified_guess": 0,
            "number_quarter_final_guess": 0,
            "number_semi_final_guess": 0,
            "number_final_guess": 0,
            "number_winner_guess": 0,
            "points": 6.0,
        },
        {
            "rank": 3,
            "first_name": users_data[2].first_name,
            "last_name": users_data[2].last_name,
            "full_name": f"{users_data[2].first_name} {users_data[2].last_name}",
            "number_match_guess": 0,
            "number_score_guess": 0,
            "number_qualified_teams_guess": 0,
            "number_first_qualified_guess": 0,
            "number_quarter_final_guess": 0,
            "number_semi_final_guess": 0,
            "number_final_guess": 0,
            "number_winner_guess": 0,
            "points": 0.0,
        },
    ]



import warnings
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from sqlalchemy import and_

from yak_server.database import build_local_session_maker
from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    MatchModel,
    ScoreBetModel,
    TeamModel,
    UserModel,
)
from yak_server.helpers.settings import get_settings

if TYPE_CHECKING:
    from sqlalchemy import Engine

try:
    import httpx
except ImportError:  # pragma: no cover
    # Very common pattern for optional dependency imports
    httpx = None  # type: ignore[assignment]

try:
    import bs4
except ImportError:  # pragma: no cover
    # Very common pattern for optional dependency imports
    bs4 = None  # type: ignore[assignment]

try:
    import lxml
except ImportError:  # pragma: no cover
    # Very common pattern for optional dependency imports
    lxml = None  # type: ignore[assignment]


def parse_score(content: str) -> tuple[Optional[int], Optional[int]]:
    try:
        scores = next(next(content.children).children)
    except AttributeError:
        return None, None

    try:
        score1_str, score2_str = scores.split("\u2013")
    except ValueError:  # pragma: no cover
        return None, None

    try:
        score1 = int(score1_str)
    except ValueError:  # pragma: no cover
        warnings.warn(f"Couldn't parse score1: '{score1_str}'", stacklevel=2)
        score1 = None

    try:
        score2 = int(score2_str)
    except ValueError:  # pragma: no cover
        warnings.warn(f"Couldn't parse score2: '{score2_str}'", stacklevel=2)
        score2 = None

    return score1, score2


def parse_penalty(content: str) -> bool:
    penalties_tag = content.find("a", string="Penalties")
    score_content = penalties_tag.next.parent.find_next("th").string

    score1, score2 = score_content.split("\u2013")

    return int(score1) > int(score2)


def is_penalties(content: str) -> bool:
    return bool(content.find("a", string="Penalties"))


class SyncOfficialResultsNotAvailableError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "Synchronize official results is not available without sync extra dependency"
            " installed. To enable it, please run: uv pip install yak-server[sync]"
        )


@dataclass
class GroupContainer:
    model: GroupModel
    content: str


@dataclass
class Group:
    index: int


@dataclass
class Team:
    score: int
    description: str
    won: Optional[bool] = None


@dataclass
class Match:
    index: int
    group: Group
    team1: Team
    team2: Team


def extract_matches_from_html(groups: list[GroupContainer]) -> list[Match]:
    matches = []

    for index, group in enumerate(groups, start=1):
        try:
            next_group = groups[index + 1]
        except IndexError:
            next_group = None

        previous_matches_html = (
            frozenset(next_group.content.find_all_previous("div", class_="footballbox"))
            if next_group is not None
            else None
        )

        next_matches_html = group.content.find_all_next("div", class_="footballbox")

        matches_html = [
            match
            for match in next_matches_html
            if previous_matches_html is None or match in previous_matches_html
        ]

        for match_index, match_html in enumerate(matches_html, start=1):
            score1, score2 = parse_score(match_html.find("th", class_="fscore"))

            match = Match(
                index=match_index,
                group=Group(index=group.model.index),
                team1=Team(
                    description=(
                        match_html.find("th", class_="fhome").get_text().replace("\xa0", "")
                    ),
                    score=score1,
                ),
                team2=Team(
                    description=(
                        match_html.find("th", class_="faway").get_text().replace("\xa0", "")
                    ),
                    score=score2,
                ),
            )

            if is_penalties(match_html):
                is_one_won = parse_penalty(match_html)

                match.team1.won = is_one_won
                match.team2.won = not is_one_won

            matches.append(match)

    return matches


def synchronize_official_results(engine: "Engine") -> None:
    if bs4 is None or lxml is None or httpx is None:
        raise SyncOfficialResultsNotAvailableError

    official_results_url = get_settings().official_results_url

    response = httpx.get(str(official_results_url))

    soup = bs4.BeautifulSoup(response.text, "lxml")

    local_session_maker = build_local_session_maker(engine)

    with local_session_maker() as db:
        groups = [
            GroupContainer(model=group, content=content.parent.parent)
            for group in db.query(GroupModel).order_by(GroupModel.index)
            for content in soup.find(
                "h3", id=group.description_en.replace(" ", "_"), string=group.description_en
            )
        ]

        matches = extract_matches_from_html(groups)

        admin = db.query(UserModel).filter_by(name="admin").first()

        for match in matches:
            score_bet = (
                db.query(ScoreBetModel)
                .join(ScoreBetModel.match)
                .join(MatchModel.group)
                .filter(
                    and_(
                        GroupModel.index == match.group.index,
                        MatchModel.index == match.index,
                        MatchModel.user_id == admin.id,
                    )
                )
                .first()
            )

            if score_bet is not None:
                score_bet.score1 = match.team1.score
                score_bet.score2 = match.team2.score

            binary_bet = (
                db.query(BinaryBetModel)
                .join(BinaryBetModel.match)
                .join(MatchModel.group)
                .join(GroupModel.phase)
                .filter(
                    and_(
                        GroupModel.index == match.group.index,
                        MatchModel.index == match.index,
                        MatchModel.user_id == admin.id,
                    )
                )
                .first()
            )

            if binary_bet is not None:
                team1 = (
                    db.query(TeamModel).filter_by(description_en=match.team1.description).first()
                )
                team2 = (
                    db.query(TeamModel).filter_by(description_en=match.team2.description).first()
                )

                if team1 is not None and team2 is not None:
                    binary_bet.match.team1_id = team1.id
                    binary_bet.match.team2_id = team2.id

                    if match.team1.won is not None:
                        binary_bet.is_one_won = match.team1.won
                    else:
                        binary_bet.is_one_won = match.team1.score > match.team2.score

        db.commit()



from dataclasses import dataclass
from http import HTTPStatus
from typing import TYPE_CHECKING, Optional

import pytest
from fastapi.testclient import TestClient
from typer.testing import CliRunner

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli import app as typer_app
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    from fastapi import FastAPI
    from sqlalchemy import Engine

runner = CliRunner()


@dataclass
class CompetitionData:
    url: str
    folder: str
    excepted_score_bets: list[tuple[Optional[int], Optional[int]]]
    excepted_binary_bets: list[Optional[bool]]


euro_2024_data = CompetitionData(
    url="https://en.wikipedia.org/w/index.php?title=UEFA_Euro_2024&oldid=1186380755",
    folder="euro_2024",
    excepted_score_bets=[
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
    ],
    excepted_binary_bets=[
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
    ],
)

world_cup_2022_data = CompetitionData(
    url="https://en.wikipedia.org/wiki/2022_FIFA_World_Cup",
    folder="world_cup_2022",
    excepted_score_bets=[
        (0, 2),
        (0, 2),
        (1, 3),
        (1, 1),
        (1, 2),
        (2, 0),
        (6, 2),
        (1, 1),
        (0, 2),
        (0, 0),
        (0, 3),
        (0, 1),
        (1, 2),
        (0, 0),
        (2, 0),
        (2, 0),
        (0, 2),
        (1, 2),
        (0, 0),
        (4, 1),
        (0, 1),
        (2, 1),
        (1, 0),
        (1, 0),
        (1, 2),
        (7, 0),
        (0, 1),
        (1, 1),
        (2, 1),
        (2, 4),
        (0, 0),
        (1, 0),
        (0, 2),
        (4, 1),
        (0, 0),
        (1, 2),
        (1, 0),
        (2, 0),
        (3, 3),
        (1, 0),
        (2, 3),
        (1, 0),
        (0, 0),
        (3, 2),
        (2, 3),
        (2, 0),
        (0, 2),
        (2, 1),
    ],
    excepted_binary_bets=[
        True,
        True,
        True,
        True,
        False,
        True,
        True,
        True,
        True,
        False,
        True,
        False,
        True,
        True,
        True,
    ],
)

euro_2020_data = CompetitionData(
    url="https://en.wikipedia.org/wiki/UEFA_Euro_2020",
    folder="euro_2020",
    excepted_score_bets=[
        (0, 3),
        (1, 1),
        (0, 2),
        (3, 0),
        (3, 1),
        (1, 0),
        (0, 1),
        (3, 0),
        (0, 1),
        (1, 2),
        (1, 4),
        (0, 2),
        (3, 1),
        (3, 2),
        (2, 1),
        (2, 0),
        (0, 3),
        (0, 1),
        (1, 0),
        (0, 2),
        (1, 1),
        (0, 0),
        (3, 1),
        (0, 1),
        (1, 2),
        (0, 0),
        (1, 0),
        (1, 1),
        (0, 5),
        (3, 2),
        (0, 3),
        (1, 0),
        (1, 1),
        (2, 4),
        (2, 2),
        (2, 2),
    ],
    excepted_binary_bets=[
        False,
        True,
        False,
        True,
        False,
        False,
        True,
        False,
        False,
        False,
        False,
        False,
        True,
        True,
        True,
    ],
)

world_cup_2018_data = CompetitionData(
    url="https://en.wikipedia.org/wiki/2018_FIFA_World_Cup",
    folder="world_cup_2018",
    excepted_score_bets=[
        (5, 0),
        (0, 1),
        (3, 1),
        (1, 0),
        (3, 0),
        (2, 1),
        (0, 1),
        (3, 3),
        (1, 0),
        (0, 1),
        (1, 1),
        (2, 2),
        (2, 1),
        (0, 1),
        (1, 1),
        (1, 0),
        (0, 0),
        (0, 2),
        (1, 1),
        (2, 0),
        (0, 3),
        (2, 0),
        (1, 2),
        (1, 2),
        (0, 1),
        (1, 1),
        (2, 0),
        (1, 2),
        (0, 2),
        (2, 2),
        (0, 1),
        (1, 0),
        (1, 2),
        (2, 1),
        (2, 0),
        (0, 3),
        (3, 0),
        (1, 2),
        (5, 2),
        (6, 1),
        (0, 1),
        (1, 2),
        (1, 2),
        (1, 2),
        (2, 2),
        (0, 3),
        (0, 1),
        (0, 1),
    ],
    excepted_binary_bets=[
        True,
        True,
        False,
        True,
        True,
        True,
        True,
        False,
        False,
        False,
        False,
        False,
        True,
        True,
        True,
    ],
)

euro_2016_data = CompetitionData(
    url="https://en.wikipedia.org/wiki/UEFA_Euro_2016",
    folder="euro_2016",
    excepted_score_bets=[
        (2, 1),
        (0, 1),
        (1, 1),
        (2, 0),
        (0, 1),
        (0, 0),
        (2, 1),
        (1, 1),
        (1, 2),
        (2, 1),
        (0, 3),
        (0, 0),
        (1, 0),
        (2, 0),
        (0, 2),
        (0, 0),
        (0, 1),
        (0, 1),
        (0, 1),
        (1, 0),
        (2, 2),
        (3, 0),
        (0, 2),
        (2, 1),
        (1, 1),
        (0, 2),
        (1, 0),
        (3, 0),
        (0, 1),
        (0, 1),
        (0, 2),
        (1, 1),
        (1, 1),
        (0, 0),
        (2, 1),
        (3, 3),
    ],
    excepted_binary_bets=[
        False,
        True,
        False,
        True,
        True,
        False,
        True,
        False,
        False,
        True,
        True,
        True,
        True,
        False,
        True,
    ],
)


def idfn(value: CompetitionData) -> str:
    return value.folder


@pytest.mark.parametrize(
    "competition_data",
    [euro_2024_data, world_cup_2022_data, euro_2020_data, world_cup_2018_data, euro_2016_data],
    ids=idfn,
)
def test_db_sync(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
    competition_data: CompetitionData,
) -> None:
    # Setup fastapi application and initialize database
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(
            data_folder_relative=f"../../yak_server/data/{competition_data.folder}",
            official_results_url=competition_data.url,
        ),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    # Signup admin user
    client = TestClient(app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": "admin",
            "first_name": "admin",
            "last_name": "admin",
            "password": get_random_string(15),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    token = response_signup.json()["result"]["token"]

    # Synchronize admin bets with official results
    monkeypatch.setattr(
        "yak_server.cli.database.sync.get_settings",
        MockSettings(official_results_url=competition_data.url),
    )

    result = runner.invoke(typer_app, ["db", "sync"])

    assert result.exit_code == 0

    # Fetch all bets and check
    response_all_bets = client.get("/api/v1/bets", headers={"Authorization": f"Bearer {token}"})

    assert response_all_bets.status_code == HTTPStatus.OK

    assert [
        (score_bet["team1"]["score"], score_bet["team2"]["score"])
        for score_bet in response_all_bets.json()["result"]["score_bets"]
    ] == competition_data.excepted_score_bets

    assert [
        binary_bet["team1"]["won"] if binary_bet["team1"] is not None else None
        for binary_bet in response_all_bets.json()["result"]["binary_bets"]
    ] == competition_data.excepted_binary_bets



"""Add group rank need_recomputation column.

Revision ID: e0f551c7766c
Revises: b36763d4cb42
Create Date: 2023-03-18 22:04:41.764896

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e0f551c7766c"
down_revision = "b36763d4cb42"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("group_position", schema=None) as batch_op:
        batch_op.add_column(sa.Column("need_recomputation", sa.Boolean(), nullable=False))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("group_position", schema=None) as batch_op:
        batch_op.drop_column("need_recomputation")

    # ### end Alembic commands ###



from collections.abc import Iterable
from dataclasses import dataclass, field
from itertools import chain
from typing import TYPE_CHECKING
from uuid import UUID

from pydantic import BaseModel
from sqlalchemy import and_

from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    MatchModel,
    PhaseModel,
    ScoreBetModel,
    UserModel,
)
from yak_server.helpers.group_position import get_group_rank_with_code

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from yak_server.database.models import GroupPositionModel


class RuleComputePoints(BaseModel):
    base_correct_result: int
    multiplying_factor_correct_result: int
    base_correct_score: int
    multiplying_factor_correct_score: int
    team_qualified: int
    first_team_qualified: int


@dataclass
class ResultForScoreBet:
    number_correct_result: int = 0
    user_ids_found_correct_result: list[UUID] = field(default_factory=list)
    number_correct_score: int = 0
    user_ids_found_correct_score: list[UUID] = field(default_factory=list)


def compute_results_for_score_bet(db: "Session", admin: UserModel) -> list[ResultForScoreBet]:
    results: list[ResultForScoreBet] = []

    for real_score in (
        db.query(ScoreBetModel).join(ScoreBetModel.match).filter(MatchModel.user_id == admin.id)
    ):
        result_for_score_bet = ResultForScoreBet()

        for user_score in (
            db.query(ScoreBetModel)
            .join(ScoreBetModel.match)
            .filter(
                and_(
                    MatchModel.user_id != admin.id,
                    MatchModel.index == real_score.match.index,
                    MatchModel.group_id == real_score.match.group_id,
                ),
            )
        ):
            if user_score.is_same_results(real_score):
                result_for_score_bet.number_correct_result += 1
                result_for_score_bet.user_ids_found_correct_result.append(user_score.match.user_id)

                if user_score.is_same_scores(real_score):
                    result_for_score_bet.number_correct_score += 1
                    result_for_score_bet.user_ids_found_correct_score.append(
                        user_score.match.user_id
                    )

        results.append(result_for_score_bet)

    return results


@dataclass
class ResultForGroupRank:
    number_qualified_teams_guess: int = 0
    number_first_qualified_guess: int = 0


def all_results_filled_in_group(group_result: list["GroupPositionModel"]) -> bool:
    return all(team.played == len(group_result) - 1 for team in group_result)


def compute_results_for_group_rank(
    db: "Session", admin: UserModel, other_users: Iterable[UserModel]
) -> dict[UUID, ResultForGroupRank]:
    result_groups: dict[UUID, ResultForGroupRank] = {}

    for group in db.query(GroupModel).join(GroupModel.phase).filter(PhaseModel.code == "GROUP"):
        group_result_admin = get_group_rank_with_code(db, admin, group.id)

        if all_results_filled_in_group(group_result_admin):
            admin_first_team_id = group_result_admin[0].team.id
            admin_second_team_id = group_result_admin[1].team.id

            for other_user in other_users:
                if other_user.id not in result_groups:
                    result_groups[other_user.id] = ResultForGroupRank()

                group_result_user = get_group_rank_with_code(db, other_user, group.id)

                if all_results_filled_in_group(group_result_user):
                    user_first_team_id = group_result_user[0].team.id
                    user_second_team_id = group_result_user[1].team.id

                    result_groups[other_user.id].number_qualified_teams_guess += len(
                        {user_first_team_id, user_second_team_id}
                        & {admin_first_team_id, admin_second_team_id},
                    )

                    if user_first_team_id == admin_first_team_id:
                        result_groups[other_user.id].number_first_qualified_guess += 1

    return result_groups


def team_from_group_code(db: "Session", user: UserModel, group_code: str) -> set[UUID]:
    return set(
        chain(
            *(
                (bet.match.team1_id, bet.match.team2_id)
                for bet in db.query(BinaryBetModel)
                .join(BinaryBetModel.match)
                .filter(and_(MatchModel.user_id == user.id, GroupModel.code == group_code))
                .join(BinaryBetModel.match)
                .join(MatchModel.group)
                if bet.match.team1_id is not None and bet.match.team2_id is not None
            ),
        ),
    )


def winner_from_user(db: "Session", user: UserModel) -> set[UUID]:
    finale_bet = next(
        iter(
            db.query(BinaryBetModel)
            .join(BinaryBetModel.match)
            .join(MatchModel.group)
            .filter(and_(MatchModel.user_id == user.id, GroupModel.code == "1"))
        ),
    )

    if (
        finale_bet.is_one_won is None
        or finale_bet.match.team1 is None
        or finale_bet.match.team2 is None
    ):
        return set()

    return {
        finale_bet.match.team1.id if finale_bet.is_one_won else finale_bet.match.team2.id,
    }


def compute_points(db: "Session", admin: UserModel, rule_config: RuleComputePoints) -> None:
    results = compute_results_for_score_bet(db, admin)

    other_users = db.query(UserModel).filter(UserModel.name != "admin")

    result_groups: dict[UUID, ResultForGroupRank] = compute_results_for_group_rank(
        db,
        admin,
        other_users,
    )

    quarter_finals_team = team_from_group_code(db, admin, "4")
    semi_finals_team = team_from_group_code(db, admin, "2")
    final_team = team_from_group_code(db, admin, "1")
    winner = winner_from_user(db, admin)

    numbers_of_players = other_users.count()

    for user in other_users:
        user.number_score_guess = 0
        user.number_match_guess = 0
        user.points = 0

        for result in results:
            if user.id in result.user_ids_found_correct_result:
                user.number_match_guess += 1
                user.points += (
                    rule_config.base_correct_result
                    + rule_config.multiplying_factor_correct_result
                    * (numbers_of_players - result.number_correct_result)
                    / (numbers_of_players - 1)
                )

            if user.id in result.user_ids_found_correct_score:
                user.number_score_guess += 1
                user.points += (
                    rule_config.base_correct_score
                    + rule_config.multiplying_factor_correct_score
                    * (numbers_of_players - result.number_correct_score)
                    / (numbers_of_players - 1)
                )

        if user.id not in result_groups:
            continue

        user.number_qualified_teams_guess = result_groups[user.id].number_qualified_teams_guess
        user.number_first_qualified_guess = result_groups[user.id].number_first_qualified_guess

        user.points += user.number_qualified_teams_guess * rule_config.team_qualified
        user.points += user.number_first_qualified_guess * rule_config.first_team_qualified

        user.number_quarter_final_guess = len(
            team_from_group_code(db, user, "4").intersection(quarter_finals_team),
        )
        user.number_semi_final_guess = len(
            team_from_group_code(db, user, "2").intersection(semi_finals_team),
        )
        user.number_final_guess = len(
            team_from_group_code(db, user, "1").intersection(final_team),
        )
        user.number_winner_guess = len(winner_from_user(db, user).intersection(winner))

        user.points += 30 * user.number_quarter_final_guess
        user.points += 60 * user.number_semi_final_guess
        user.points += 120 * user.number_final_guess
        user.points += 200 * user.number_winner_guess

    db.commit()



from pydantic import UUID4, BaseModel, ConfigDict


class SignupIn(BaseModel):
    name: str
    first_name: str
    last_name: str
    password: str

    model_config = ConfigDict(extra="forbid")


class SignupOut(BaseModel):
    id: UUID4
    name: str
    token: str

    model_config = ConfigDict(extra="forbid")


class PasswordRequirementsOut(BaseModel):
    minimum_length: int
    uppercase: bool
    lowercase: bool
    digit: bool
    no_space: bool

    model_config = ConfigDict(extra="forbid")


class LoginIn(BaseModel):
    name: str
    password: str

    model_config = ConfigDict(extra="forbid")


class LoginOut(BaseModel):
    id: UUID4
    name: str
    token: str

    model_config = ConfigDict(extra="forbid")


class CurrentUserOut(BaseModel):
    id: UUID4
    name: str

    model_config = ConfigDict(from_attributes=True)


class ModifyUserIn(BaseModel):
    password: str

    model_config = ConfigDict(extra="forbid")



from secrets import SystemRandom, randbelow
from typing import TYPE_CHECKING

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_modify_score_bet_and_group_rank(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    client = TestClient(app_with_valid_jwt_config)

    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_modify_score_bet_and_group_rank"),
    )
    initialize_database(engine_for_test, app_with_valid_jwt_config)

    query_signup = """
        mutation Root(
            $userName: String!, $firstName: String!,
            $lastName: String!, $password: String!
        ) {
            signupResult(
                userName: $userName, firstName: $firstName,
                lastName: $lastName, password: $password
            ) {
                __typename
                ... on UserWithToken {
                    token
                    scoreBets {
                        id
                        team1 {
                            score
                            description
                        }
                        team2 {
                            score
                            description
                        }
                    }
                }
                ... on UserNameAlreadyExists {
                    message
                }
            }
        }
    """

    response_signup = client.post(
        "/api/v2",
        json={
            "query": query_signup,
            "variables": {
                "userName": get_random_string(6),
                "firstName": get_random_string(6),
                "lastName": get_random_string(6),
                "password": get_random_string(15),
            },
        },
    )

    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"
    token = response_signup.json()["data"]["signupResult"]["token"]

    # Success case : Modify score bet and observe group rank update
    query_modify_score_and_group_rank = """
        mutation Root($id: UUID!, $score1: Int, $score2: Int) {
            modifyScoreBetResult(id: $id, score1: $score1, score2: $score2) {
                __typename
                ... on ScoreBet {
                    id
                    locked
                    group {
                        description
                        groupRank {
                            team {
                                description
                            }
                            won
                            drawn
                            lost
                            points
                            goalsFor
                            goalsAgainst
                            goalsDifference
                        }
                        phase {
                            description
                        }
                    }
                    team1 {
                        score
                        description
                    }
                    team2 {
                        score
                        description
                    }
                }
            }
        }
    """

    score1 = SystemRandom().randrange(1, 7)
    score2 = randbelow(score1)

    # Apply twice the same update to check first, updates are ok and then second, that nothing
    # change
    for _ in range(2):
        response_modify_score_bet = client.post(
            "/api/v2",
            headers={"Authorization": f"Bearer {token}"},
            json={
                "query": query_modify_score_and_group_rank,
                "variables": {
                    "id": response_signup.json()["data"]["signupResult"]["scoreBets"][0]["id"],
                    "score1": score1,
                    "score2": score2,
                },
            },
        )

        group_rank = response_modify_score_bet.json()["data"]["modifyScoreBetResult"]["group"][
            "groupRank"
        ]

        assert group_rank[0]["goalsFor"] == score1
        assert group_rank[0]["goalsAgainst"] == score2
        assert group_rank[3]["goalsFor"] == score2
        assert group_rank[3]["goalsAgainst"] == score1



from typing import Annotated, Union
from uuid import UUID

import strawberry

from yak_server.helpers.errors import (
    EXPIRED_TOKEN_MESSAGE,
    INVALID_CREDENTIALS_MESSAGE,
    INVALID_TOKEN_MESSAGE,
    LOCKED_BINARY_BET_MESSAGE,
    LOCKED_SCORE_BET_MESSAGE,
    UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE,
    binary_bet_not_found_message,
    group_not_found_message,
    name_already_exists_message,
    phase_not_found_message,
    score_bet_not_found_message,
    team_not_found_message,
    user_not_found_message,
)

from .schema import (
    BinaryBet,
    Group,
    GroupPosition,
    Phase,
    ScoreBet,
    Team,
    User,
    UserWithoutSensitiveInfo,
    UserWithToken,
)


@strawberry.type
class InvalidToken:
    message: str = INVALID_TOKEN_MESSAGE


@strawberry.type
class ExpiredToken:
    message: str = EXPIRED_TOKEN_MESSAGE


@strawberry.type
class UnauthorizedAccessToAdminAPI:
    message: str = UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE


@strawberry.type
class LockedScoreBetError:
    message: str = LOCKED_SCORE_BET_MESSAGE


@strawberry.type
class ScoreBetNotFoundForUpdate:
    message: str = "Score bet not found. Cannot modify a resource that does not exist."


@strawberry.type
class NewScoreNegative:
    variable_name: strawberry.Private[str]
    score: strawberry.Private[int]

    @strawberry.field
    def message(self) -> str:
        return (
            f"Variable '{self.variable_name}' got invalid value {self.score}. "
            "Score cannot be negative."
        )


ModifyScoreBetResult = Annotated[
    Union[
        ScoreBet,
        ScoreBetNotFoundForUpdate,
        LockedScoreBetError,
        NewScoreNegative,
        InvalidToken,
        ExpiredToken,
    ],
    strawberry.union("ModifyScoreBetResult"),
]


@strawberry.type
class LockedBinaryBetError:
    message: str = LOCKED_BINARY_BET_MESSAGE


@strawberry.type
class BinaryBetNotFoundForUpdate:
    message: str = "Binary bet not found. Cannot modify a resource that does not exist."


ModifyBinaryBetResult = Annotated[
    Union[
        BinaryBet,
        BinaryBetNotFoundForUpdate,
        LockedBinaryBetError,
        InvalidToken,
        ExpiredToken,
    ],
    strawberry.union("ModifyBinaryBetResult"),
]


CurrentUserResult = Annotated[
    Union[User, InvalidToken, ExpiredToken],
    strawberry.union("CurrentUserResult"),
]


@strawberry.type
class AllTeamsSuccessful:
    teams: list[Team]


AllTeamsResult = Annotated[
    Union[AllTeamsSuccessful, InvalidToken, ExpiredToken],
    strawberry.union("AllTeamsResult"),
]


@strawberry.type
class TeamByIdNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return team_not_found_message(self.id)


@strawberry.type
class TeamByCodeNotFound:
    code: strawberry.Private[str]

    @strawberry.field
    def message(self) -> str:
        return team_not_found_message(self.code)


TeamByIdResult = Annotated[
    Union[Team, TeamByIdNotFound, InvalidToken, ExpiredToken],
    strawberry.union("TeamByIdResult"),
]

TeamByCodeResult = Annotated[
    Union[Team, TeamByCodeNotFound, InvalidToken, ExpiredToken],
    strawberry.union("TeamByCodeResult"),
]


@strawberry.type
class ScoreBetNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return score_bet_not_found_message(self.id)


ScoreBetResult = Annotated[
    Union[ScoreBet, ScoreBetNotFound, InvalidToken, ExpiredToken],
    strawberry.union("ScoreBetResult"),
]


@strawberry.type
class BinaryBetNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return binary_bet_not_found_message(self.id)


BinaryBetResult = Annotated[
    Union[
        BinaryBet,
        BinaryBetNotFound,
        InvalidToken,
        ExpiredToken,
    ],
    strawberry.union("BinaryBetResult"),
]


@strawberry.type
class Groups:
    groups: list[Group]


AllGroupsResult = Annotated[
    Union[Groups, InvalidToken, ExpiredToken],
    strawberry.union("AllGroupsResult"),
]


@strawberry.type
class GroupByCodeNotFound:
    code: strawberry.Private[str]

    @strawberry.field
    def message(self) -> str:
        return group_not_found_message(self.code)


GroupByCodeResult = Annotated[
    Union[Group, GroupByCodeNotFound, InvalidToken, ExpiredToken],
    strawberry.union(
        "GroupByCodeResult",
    ),
]


@strawberry.type
class GroupByIdNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return group_not_found_message(self.id)


GroupByIdResult = Annotated[
    Union[Group, GroupByIdNotFound, InvalidToken, ExpiredToken],
    strawberry.union(
        "GroupByIdResult",
    ),
]


@strawberry.type
class Phases:
    phases: list[Phase]


AllPhasesResult = Annotated[
    Union[Phases, InvalidToken, ExpiredToken],
    strawberry.union("AllPhasesResult"),
]


@strawberry.type
class PhaseByCodeNotFound:
    code: strawberry.Private[str]

    @strawberry.field
    def message(self) -> str:
        return phase_not_found_message(self.code)


PhaseByCodeResult = Annotated[
    Union[Phase, PhaseByCodeNotFound, InvalidToken, ExpiredToken],
    strawberry.union(
        "PhaseByCodeResult",
    ),
]


@strawberry.type
class PhaseByIdNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return phase_not_found_message(self.id)


PhaseByIdResult = Annotated[
    Union[Phase, PhaseByIdNotFound, InvalidToken, ExpiredToken],
    strawberry.union("PhaseByIdResult"),
]


@strawberry.type
class ScoreBoard:
    users: list[UserWithoutSensitiveInfo]


ScoreBoardResult = Annotated[
    Union[ScoreBoard, InvalidToken, ExpiredToken],
    strawberry.union(
        "ScoreBoardResult",
    ),
]


@strawberry.type
class UserNameAlreadyExists:
    user_name: strawberry.Private[str]

    @strawberry.field
    def message(self) -> str:
        return name_already_exists_message(self.user_name)


@strawberry.type
class UnsatisfiedPasswordRequirements:
    message: str


SignupResult = Annotated[
    Union[UserWithToken, UserNameAlreadyExists, UnsatisfiedPasswordRequirements],
    strawberry.union("SignupResult"),
]


@strawberry.type
class InvalidCredentials:
    message: str = INVALID_CREDENTIALS_MESSAGE


LoginResult = Annotated[Union[UserWithToken, InvalidCredentials], strawberry.union("LoginResult")]


@strawberry.type
class GroupRank:
    group_rank: list[GroupPosition]
    group: Group


GroupRankByCodeResult = Annotated[
    Union[GroupRank, GroupByCodeNotFound, InvalidToken, ExpiredToken],
    strawberry.union(
        "GroupRankByCodeResult",
    ),
]


GroupRankByIdResult = Annotated[
    Union[GroupRank, GroupByIdNotFound, InvalidToken, ExpiredToken],
    strawberry.union(
        "GroupRankByIdResult",
    ),
]


@strawberry.type
class UserNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return user_not_found_message(self.id)


ModifyUserResult = Annotated[
    Union[
        UserWithoutSensitiveInfo,
        UserNotFound,
        InvalidToken,
        ExpiredToken,
        UnauthorizedAccessToAdminAPI,
    ],
    strawberry.union(
        "ModifyUserResult",
    ),
]



from http import HTTPStatus
from typing import TYPE_CHECKING

import pendulum
from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database
from yak_server.helpers.settings import get_settings

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_delete_binary_bet(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_binary_bet"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    # Signup one user
    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(6),
            "first_name": get_random_string(6),
            "last_name": get_random_string(6),
            "password": get_random_string(16),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    # Fetch all bets and get the binary bet id
    token = response_signup.json()["result"]["token"]

    response_all_bets = client.get("/api/v1/bets", headers={"Authorization": f"Bearer {token}"})

    assert response_all_bets.status_code == HTTPStatus.OK

    binary_bet_id = response_all_bets.json()["result"]["binary_bets"][0]["id"]

    # Check bet locking
    app_with_valid_jwt_config.dependency_overrides[get_settings]().set_lock_datetime(
        -pendulum.duration(minutes=10)
    )

    response_delete_locked_binary_bet = client.delete(
        f"/api/v1/binary_bets/{binary_bet_id}",
        headers={"Authorization": f"Bearer {token}"},
    )

    assert response_delete_locked_binary_bet.status_code == HTTPStatus.UNAUTHORIZED
    assert response_delete_locked_binary_bet.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Cannot modify binary bet, lock date is exceeded",
    }

    app_with_valid_jwt_config.dependency_overrides[get_settings]().set_lock_datetime(
        pendulum.duration(minutes=10)
    )

    # Retrieve one binary bet
    response_get_binary_bet = client.get(
        f"/api/v1/binary_bets/{binary_bet_id}",
        headers={"Authorization": f"Bearer {token}"},
    )

    assert response_get_binary_bet.status_code == HTTPStatus.OK

    # Delete one binary bet and compare it to the GET response
    response_delete_binary_bet = client.delete(
        f"/api/v1/binary_bets/{binary_bet_id}",
        headers={"Authorization": f"Bearer {token}"},
    )

    assert response_delete_binary_bet.status_code == HTTPStatus.OK
    assert response_delete_binary_bet.json() == response_get_binary_bet.json()

    # Try to delete a second times and error
    response_delete_binary_bet_second_times = client.delete(
        f"/api/v1/binary_bets/{binary_bet_id}",
        headers={"Authorization": f"Bearer {token}"},
    )

    assert response_delete_binary_bet_second_times.status_code == HTTPStatus.NOT_FOUND
    assert response_delete_binary_bet_second_times.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Bet not found: {binary_bet_id}",
    }



import logging
import traceback
from typing import TYPE_CHECKING, Union
from uuid import UUID

from fastapi import HTTPException, Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from starlette.exceptions import HTTPException as StarletteHTTPException

from yak_server.helpers.errors import (
    EXPIRED_TOKEN_MESSAGE,
    INVALID_CREDENTIALS_MESSAGE,
    INVALID_TOKEN_MESSAGE,
    LOCKED_BINARY_BET_MESSAGE,
    LOCKED_SCORE_BET_MESSAGE,
    UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE,
    bet_not_found_message,
    group_not_found_message,
    name_already_exists_message,
    phase_not_found_message,
    rule_not_found_message,
    team_not_found_message,
    user_not_found_message,
)

if TYPE_CHECKING:
    from fastapi import FastAPI

logger = logging.getLogger(__name__)


class InvalidCredentials(HTTPException):
    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=INVALID_CREDENTIALS_MESSAGE,
        )


class NameAlreadyExists(HTTPException):
    def __init__(self, name: str) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=name_already_exists_message(name),
        )


class BetNotFound(HTTPException):
    def __init__(self, bet_id: UUID) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=bet_not_found_message(bet_id),
        )


class UnsatisfiedPasswordRequirements(HTTPException):
    def __init__(self, detail: str) -> None:
        super().__init__(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Unsatisfied password requirements. {detail}",
        )


class UserNotFound(HTTPException):
    def __init__(self, user_id: UUID) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=user_not_found_message(user_id),
        )


class UnauthorizedAccessToAdminAPI(HTTPException):
    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE,
        )


class NoAdminUser(HTTPException):
    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="No admin user found",
        )


class InvalidTeamId(HTTPException):
    def __init__(self, team_id: str) -> None:
        super().__init__(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Invalid team id: {team_id}. Retry with a uuid or ISO 3166-1 alpha-2 code",
        )


class TeamNotFound(HTTPException):
    def __init__(self, team_id: Union[str, UUID]) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=team_not_found_message(team_id),
        )


class LockedScoreBet(HTTPException):
    def __init__(self) -> None:
        super().__init__(status_code=status.HTTP_401_UNAUTHORIZED, detail=LOCKED_SCORE_BET_MESSAGE)


class LockedBinaryBet(HTTPException):
    def __init__(self) -> None:
        super().__init__(status_code=status.HTTP_401_UNAUTHORIZED, detail=LOCKED_BINARY_BET_MESSAGE)


class NoResultsForAdminUser(HTTPException):
    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="No results for admin user",
        )


class GroupNotFound(HTTPException):
    def __init__(self, group_id: Union[str, UUID]) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=group_not_found_message(group_id),
        )


class PhaseNotFound(HTTPException):
    def __init__(self, phase_id: Union[str, UUID]) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=phase_not_found_message(phase_id),
        )


class InvalidToken(HTTPException):
    def __init__(self) -> None:
        super().__init__(status_code=status.HTTP_401_UNAUTHORIZED, detail=INVALID_TOKEN_MESSAGE)


class ExpiredToken(HTTPException):
    def __init__(self) -> None:
        super().__init__(status_code=status.HTTP_401_UNAUTHORIZED, detail=EXPIRED_TOKEN_MESSAGE)


class RuleNotFound(HTTPException):
    def __init__(self, rule_id: UUID) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=rule_not_found_message(rule_id),
        )


def log_traceback(exception: Exception) -> None:
    """
    Log the traceback of an exception.
    """
    for frame in traceback.format_list(traceback.extract_tb(exception.__traceback__)):
        logger.info(frame)

    if exception.__cause__:
        for frame in traceback.format_list(traceback.extract_tb(exception.__cause__.__traceback__)):
            logger.info(frame)


def set_exception_handler(app: "FastAPI") -> None:
    @app.exception_handler(StarletteHTTPException)
    def http_exception_handler(_: Request, http_exception: StarletteHTTPException) -> JSONResponse:
        log_traceback(http_exception)

        logger.info(
            f"An expected exception occurs: {type(http_exception).__name__} {http_exception}"
        )

        return JSONResponse(
            status_code=http_exception.status_code,
            content={
                "ok": False,
                "error_code": http_exception.status_code,
                "description": http_exception.detail,
            },
        )

    @app.exception_handler(Exception)
    def handle_exception(_: Request, exception: Exception) -> JSONResponse:  # pragma: no cover
        # Return JSON instead of HTML for generic errors.
        logger.error(traceback.format_exc())
        logger.error(f"An unexpected exception occurs: {type(exception).__name__} {exception}")

        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            content={
                "ok": False,
                "error_code": status.HTTP_500_INTERNAL_SERVER_ERROR,
                "description": (
                    f"{type(exception).__name__}: {exception!s}"
                    if app.debug
                    else "Unexpected error"
                ),
            },
        )

    @app.exception_handler(RequestValidationError)
    def request_validator_error_handler(
        _: Request,
        request_validator_error: RequestValidationError,
    ) -> JSONResponse:
        log_traceback(request_validator_error)

        logger.info(
            "An validation error occurs: "
            f"{type(request_validator_error).__name__} {request_validator_error}"
        )

        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            content={
                "ok": False,
                "error_code": status.HTTP_422_UNPROCESSABLE_ENTITY,
                "description": request_validator_error.errors(),
            },
        )



from collections.abc import Awaitable
from pathlib import Path
from typing import TYPE_CHECKING, Callable
from uuid import uuid4

from fastapi import FastAPI

from yak_server import create_app as create_yak_app

try:
    import yappi
except ImportError:  # pragma: no cover
    # Very common pattern for optional dependency imports
    yappi = None  # type: ignore[assignment]

if TYPE_CHECKING:
    from fastapi import Request, Response


def set_yappi_profiler(app: "FastAPI") -> None:
    if yappi is None:
        msg = "Profiling is not available without yappi installed."
        raise NotImplementedError(msg)

    @app.middleware("http")
    async def profile_process_time(
        request: "Request",
        call_next: Callable[["Request"], Awaitable["Response"]],
    ) -> "Response":
        yappi.clear_stats()
        yappi.set_clock_type("cpu")
        yappi.start()

        response = await call_next(request)

        yappi.stop()

        folder_location = Path(__file__).parents[1] / "profiling"
        folder_location.mkdir(parents=True, exist_ok=True)

        profiling_log_id = uuid4()

        yappi.convert2pstats(yappi.get_func_stats()).dump_stats(
            folder_location / f"{profiling_log_id}.pstats"
        )

        response.headers["profiling-log-id"] = str(profiling_log_id)

        return response


def create_app() -> FastAPI:
    app = create_yak_app()

    set_yappi_profiler(app)

    return app



from typing import TYPE_CHECKING

import pytest

from testing.mock import MockSettings
from yak_server import create_app
from yak_server.cli.database import (
    MissingPhaseDuringInitError,
    MissingTeamDuringInitError,
    initialize_database,
)

if TYPE_CHECKING:
    from sqlalchemy import Engine


def test_missing_phase(monkeypatch: pytest.MonkeyPatch, engine_for_test: "Engine") -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_missing_phase"),
    )

    app = create_app()

    with pytest.raises(MissingPhaseDuringInitError) as exception:
        initialize_database(engine_for_test, app)

    assert (
        str(exception.value) == "Error during database initialization: phase_code=GROUP not found."
    )


def test_missing_team1(monkeypatch: pytest.MonkeyPatch, engine_for_test: "Engine") -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_missing_team1"),
    )

    app = create_app()

    with pytest.raises(MissingTeamDuringInitError) as exception:
        initialize_database(engine_for_test, app)

    assert str(exception.value) == "Error during database initialization: team_code=AD not found."


def test_missing_team2(monkeypatch: pytest.MonkeyPatch, engine_for_test: "Engine") -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_missing_team2"),
    )

    app = create_app()

    with pytest.raises(MissingTeamDuringInitError) as exception:
        initialize_database(engine_for_test, app)

    assert str(exception.value) == "Error during database initialization: team_code=BR not found."



from secrets import SystemRandom
from typing import TYPE_CHECKING
from uuid import uuid4

import pendulum
from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database
from yak_server.helpers.settings import get_settings

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_modify_score_bet(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    client = TestClient(app_with_valid_jwt_config)

    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_modify_bet_v2"),
    )
    initialize_database(engine_for_test, app_with_valid_jwt_config)

    user_name = get_random_string(10)
    first_name = get_random_string(5)
    last_name = get_random_string(8)
    password = get_random_string(15)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": """
                mutation Signup(
                    $userName: String!, $firstName: String!,
                    $lastName: String!, $password: String!
                ) {
                    signupResult(
                        userName: $userName, firstName: $firstName,
                        lastName: $lastName, password: $password
                    ) {
                        __typename
                        ... on UserWithToken {
                            firstName
                            lastName
                            fullName
                            token
                            scoreBets {
                                id
                            }
                        }
                        ... on UserNameAlreadyExists {
                            message
                        }
                    }
                }
            """,
            "variables": {
                "userName": user_name,
                "firstName": first_name,
                "lastName": last_name,
                "password": password,
            },
        },
    )

    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"
    score_bet_ids = [
        score_bet["id"] for score_bet in response_signup.json()["data"]["signupResult"]["scoreBets"]
    ]
    authentication_token = response_signup.json()["data"]["signupResult"]["token"]

    # Success case : check update is OK
    score1 = 4
    score2 = 4

    query_modify_score_bet = """
        mutation ModifyScoreBet($id: UUID!, $score1: Int, $score2: Int) {
            modifyScoreBetResult(id: $id, score1: $score1, score2: $score2) {
                __typename
                ... on ScoreBet {
                    id
                    team1 {
                        id
                        description
                        score
                        flag {
                            url
                        }
                    }
                    team2 {
                        description
                        score
                        flag {
                            url
                        }
                    }
                    group {
                        id
                        description
                        phase {
                            description
                        }
                    }
                }
                ... on ScoreBetNotFoundForUpdate {
                    message
                }
                ... on LockedScoreBetError {
                    message
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on NewScoreNegative {
                    message
                }
            }
        }
    """

    response_modify_bet = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={
            "query": query_modify_score_bet,
            "variables": {
                "id": score_bet_ids[0],
                "score1": score1,
                "score2": score2,
            },
        },
    )

    new_bet = response_modify_bet.json()["data"]["modifyScoreBetResult"]

    assert new_bet["__typename"] == "ScoreBet"
    assert new_bet["team1"]["score"] == score1
    assert new_bet["team2"]["score"] == score2
    assert new_bet["team1"]["flag"]["url"] == f"/api/v1/teams/{new_bet['team1']['id']}/flag"

    # Error case : authentication token is invalid
    response_modify_bet_auth_error = client.post(
        "/api/v2",
        headers={"Authorization": "InvalidKey llllllllllllllll"},
        json={
            "query": query_modify_score_bet,
            "variables": {
                "id": score_bet_ids[0],
                "score1": score1,
                "score2": score2,
            },
        },
    )

    assert response_modify_bet_auth_error.json() == {
        "data": {
            "modifyScoreBetResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }

    # Error case : check NewScoreNegative error is send back if one of score is negative
    score1 = 5
    score2 = -1

    response_modify_bet_new_score2_negative = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={
            "query": query_modify_score_bet,
            "variables": {
                "id": score_bet_ids[0],
                "score1": score1,
                "score2": score2,
            },
        },
    )

    assert response_modify_bet_new_score2_negative.json()["data"]["modifyScoreBetResult"] == {
        "__typename": "NewScoreNegative",
        "message": f"Variable '$score2' got invalid value {score2}. Score cannot be negative.",
    }

    # Error case : check NewScoreNegative error is send back if one of score1 is negative
    score1 = -SystemRandom().randrange(2, 9)
    score2 = 0

    response_modify_bet_new_score1_negative = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={
            "query": query_modify_score_bet,
            "variables": {
                "id": score_bet_ids[0],
                "score1": score1,
                "score2": score2,
            },
        },
    )

    assert response_modify_bet_new_score1_negative.json()["data"]["modifyScoreBetResult"] == {
        "__typename": "NewScoreNegative",
        "message": f"Variable '$score1' got invalid value {score1}. Score cannot be negative.",
    }

    # Error case : check ScoreBetNotFoundForUpdate error if score bet does not exist
    score1 = 1
    score2 = 1

    response_modify_bet_invalid_id = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={
            "query": query_modify_score_bet,
            "variables": {
                "id": str(uuid4()),
                "score1": score1,
                "score2": score2,
            },
        },
    )

    assert response_modify_bet_invalid_id.json()["data"]["modifyScoreBetResult"] == {
        "__typename": "ScoreBetNotFoundForUpdate",
        "message": "Score bet not found. Cannot modify a resource that does not exist.",
    }

    # Error case : check locked score bet
    app_with_valid_jwt_config.dependency_overrides[get_settings] = MockSettings(
        jwt_expiration_time=100,
        jwt_secret_key=app_with_valid_jwt_config.dependency_overrides[
            get_settings
        ]().jwt_secret_key,
        lock_datetime_shift=pendulum.duration(seconds=-10),
    )

    response_modify_locked_score_bet = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={
            "query": query_modify_score_bet,
            "variables": {
                "id": score_bet_ids[1],
                "score1": 1,
                "score2": 1,
            },
        },
    )

    assert response_modify_locked_score_bet.json() == {
        "data": {
            "modifyScoreBetResult": {
                "__typename": "LockedScoreBetError",
                "message": "Cannot modify score bet, lock date is exceeded",
            },
        },
    }



from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict

Result = TypeVar("Result")


class GenericOut(BaseModel, Generic[Result]):
    ok: bool = True
    result: Result

    model_config = ConfigDict(extra="forbid")
